from __future__ import absolute_import, division, print_function

from .Patterns import Pattern, PGroup, asStream
from .TimeVar import TimeVar
from .Root import Root

from random import choice
from copy import copy
import math

def miditofreq(midinote):
    """ Converts a midi number to frequency """
    return 440 * (2 ** ((midinote - 69.0)/12.0))

def _log2(num):
    return math.log(num) / math.log(2)

def freqtomidi(freq):
    return 12 * _log2((freq / 440)) + 69

def midi(scale, octave, degree, root=0, stepsPerOctave=12):
    """ Calculates a midinote from a scale, octave, degree, and root """

    # Make sure we force timevars into real values

    if isinstance(scale, ScalePattern) and isinstance(scale.data, TimeVar):

        scale = asStream(scale.data.now())

    # Force float
    octave = float(octave)
    degree = float(degree)
    root   = float(root)
    
    # Floor val
    lo = int(math.floor(degree))
    hi = lo + 1

    octave = octave + (lo // len(scale))
    index  = lo % len(scale)

    # Work out any microtones

    micro = (degree - lo)

    if micro > 0:

        ex_scale = list(scale) + [stepsPerOctave]

        diff  = ex_scale[index + 1] - scale[index]

        micro = micro * diff

    midival = stepsPerOctave * octave # Root note of scale
    midival = midival + root          # Adjust for key
    midival = midival + scale[index]  # Add the note
    midival = midival + micro         # And any microtonal

    return midival

def get_freq_and_midi(degree, octave, root, scale):
    """ Returns the frequency and midinote """

    # TODO -- make sure it's always a scale
    if hasattr(scale, "now"):

        scale = scale.now()

    if isinstance(scale, ScaleType):

        freq, midinote = scale.get_freq(degree, octave, root, get_midi=True)

    else:

        midinote = midi(scale, octave, degree, root)
        freq     = miditofreq(midinote)

    return freq, midinote

class ScaleType:
    pass

class TuningType(list):
    def __init__(self, data):
        data = list(data)
        list.__init__(self, data[:-1])
        self.steps = int(data[-1])

class Tuning:
    ET12          = TuningType([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12])
    just          = TuningType([0.0, 1.1173128526978, 2.0391000173077, 3.1564128700055, 3.8631371386483, 4.9804499913461, 5.9022371559561, 7.0195500086539, 8.1368628613517, 8.8435871299945, 10.175962878659, 10.882687147302, 12 ])
    bohlen_pierce = TuningType([i*12/13*math.log(3, 2) for i in range(14)])

class ScalePattern(ScaleType, Pattern):

    name = None

    def __init__(self, semitones, name=None, tuning=Tuning.ET12):

        self.name = name

        self.semitones = semitones.data if isinstance(semitones, Pattern) else semitones

        if not isinstance(tuning, TuningType):

            self.tuning = TuningType(tuning)

        else:

            self.tuning = tuning

        self.data = self.semitones

        self.steps = self.tuning.steps

        if self.steps:

            self.pentatonic = PentatonicScalePattern(self)

    def __eq__(self, other):
        return self.name == other.name if isinstance(other, ScalePattern) else False

    def __ne__(self, other):
        return self.name != other.name if isinstance(other, ScalePattern) else True

    def semitones(self, pitches):
        """ Returns the semitone values for a series of pitches in this scale """
        tones = []
        for pitch in asStream(pitches):
            tones.append( self.note_to_semitone(pitch) )
        return Pattern(tones)

    def get_tuned_note(self, degree):
        tuning_index = int(self[degree]) % len(self.tuning)
        # tuning_offset = 0
        # if degree < 0:
        #     tuning_offset = (((abs(degree) // len(self.tuning)) + 1) * self.steps)
        return self.tuning[tuning_index]

    def get_midi_note(self, degree, octave=5, root=0):
        """ Calculates a midinote from a scale, octave, degree, and root """

        # Make sure we force timevars into real values

        if isinstance(self.data, TimeVar):

            scale = asStream(self.data.now())

        # Force float
        octave = float(octave)
        degree = float(degree)
        root   = float(root)
        
        # Floor val
        lo = int(math.floor(degree))
        hi = lo + 1

        octave = octave + (lo // len(self))
        index  = lo % len(self)

        pitch = self.get_tuned_note(index)

        # Work out any microtones

        micro = (degree - lo)

        if micro > 0:

            ex_scale = list(self) + [self.steps]

            diff  = ex_scale[index + 1] - self[index]

            micro = micro * diff

        midival = self.steps * octave    # Root note of scale
        midival = midival + root         # Adjust for key
        midival = midival + pitch        # Add the note
        midival = midival + micro        # And any microtones
        
        return midival

    def get_freq(self, degree, octave=5, root=0, get_midi=False):
        """ Returns the frequency of a midinote calculated by self.get_midi_note. Returns a tuple containing
            the freqency and midinote if `get_midi` is set to `True`. """
        midinote = self.get_midi_note(degree, octave, root)
        return (miditofreq(midinote), midinote) if get_midi else miditofreq(midinote)

    def note_to_semitone(self, pitch):
        """ Takes a pitch value and returns the semitone value e.g. midinote value not accounting for octaves """
        if isinstance(pitch, PGroup):
            return pitch.__class__([self.note_to_semitone(p) for p in pitch])
        else:
            i = pitch % len(self.data)
            n = (pitch // len(self.data)) * self.steps 
        return asStream(self.data)[i] + n

    def getslice(self, start, stop, step=1):
        """ Called when using __getitem__ with slice notation. Numbers 
            smaller than 0 and greater than the max value are adjusted. """
        
        start = start if start is not None else 0
        stop  = stop if stop is not None else len(self)
        step  = step if step is not None else 1

        if stop < start:

            stop = (len(self.data) +  stop)

        semitones = []

        for i in range(start, stop, step):

            # Get the semitone

            tone = self[i]

            # Negative values

            if i < 0:

                sub = (((abs(i) // len(self)) + 1) * self.steps)

                tone -= sub

            # Values past the end

            elif i >= len(self):

                add = ((i // len(self)) * self.steps)

                tone += add

            semitones.append(tone)

        return ScalePattern(semitones)

    #def semitone_to_note(self, semitone):
    #    """ Takes a semitone value (midinote) and returns the pitch in this scale """
    #    return semitone


class PentatonicScalePattern(ScalePattern):

    def __init__(self, scale):

        self.update(scale)

    def update(self, scale):

        self.data      = scale
        self.steps     = scale.steps
        self.semitones = scale.semitones
        self.tuning    = scale.tuning

    def __len__(self):
        return 5

    @staticmethod
    def values(scale):

        shift = ((scale[2]-scale[0]) % 2) * 2

        return sorted([scale[(i+shift)%len(scale)] for i in range(0,5*4,4)])

    def __str__(self):
        return str(self.values(self.data))

    def __repr__(self):
        return str(self)

    def __iter__(self):

        for note in self.values(self.data):

            yield note

    def __getitem__(self, key):
        return self.values(self.data)[int(key)]

class FreqScalePattern(ScalePattern):
    def __init__(self):
        ScalePattern.__init__(self, [], name="freq")
    def get_midi_note(self, freq, *args, **kwargs):
        return freqtomidi(freq)
    def get_freq(self, freq, *args, **kwargs):
        return (freq, freqtomidi(freq)) if kwargs.get("get_midi", False) else freq
    def __repr__(self):
        return "[inf]"

class _DefaultTuning(TuningType):
    """ Wrapper for Tuning.default """
    def __init__(self, tuning):
        self.tuning = tuning

    def __len__(self):
        return len(self.tuning)

class _DefaultScale(ScaleType):
    """ Wrapper for Scale.default """
    def __init__(self, scale):
        self.scale = copy(scale)
        self.pentatonic = copy(self.scale.pentatonic)

    def __len__(self):
        return len(self.scale)

    def __repr__(self):
        return repr(self.scale)

    def __iter__(self):
        return self.scale.__iter__()

    def set(self, new, *args, **kwargs):
        """ Change the contents of the default scale """

        if type(new) == str:

            self.scale = Scale.get_scale(new)

            if "tuning" in kwargs:

                self.scale.tuning = kwargs["tuning"]

            self.pentatonic.update(self.scale.pentatonic)

        elif isinstance(new, (list, Pattern, TimeVar)):

            self.scale = ScalePattern(new, *args, **kwargs)

            # Store if the user has used a name

            if self.scale.name is not None and self.scale.name not in Scale.names():

                Scale[self.scale.name] = self.scale

            self.pentatonic.update(self.scale.pentatonic)

        else:

            print("Warning: {!r} is not a valid scale".format(new))

        return self

    def __getattribute__(self, attr):
        if attr not in ("scale", "set", "pentatonic"):
            return self.scale.__getattribute__(attr)
        else:
            return object.__getattribute__(self, attr)

    # Python2
    def __getattr__(self, attr):
        return self.__getattribute__(attr)

# Custom made fibonacci tuing

##fib = [0,1]
##for n in range(2,11):
##    fib.append(fib[n-1]+fib[n-2])
##
##fibonacci = []
##for n in range(3, len(fib)-1):
##    fibonacci.append((n-3) * (fib[n] / float(fib[n-1])))
##
##fibonacci = Scale("fibonacci", fibonacci)
##
##del n

class __scale__:

    chromatic       = ScalePattern([0,1,2,3,4,5,6,7,8,9,10,11], name="chromatic")

    major           = ScalePattern([0,2,4,5,7,9,11], name="major")
    majorPentatonic = ScalePattern([0,2,4,7,9], name="majorPentatonic" )

    minor           = ScalePattern([0,2,3,5,7,8,10], name="minor")
    aeolian         = ScalePattern([0,2,3,5,7,8,10], name="aeolian")
    minorPentatonic = ScalePattern([0,3,5,7,10], name="minorPentatonic")

    mixolydian      = ScalePattern([0,2,4,5,7,9,10], name="mixolydian")

    melodicMinor    = ScalePattern([0,2,3,5,7,9,11], name="melodicMinor")
    melodicMajor    = ScalePattern([0,2,4,5,7,8,11], name="melodicMajor")

    harmonicMinor   = ScalePattern([0,2,3,5,7,8,11], name="harmonicMinor")
    harmonicMajor   = ScalePattern([0,2,4,5,7,8,11], name="harmonicMajor")

    # Goal: Do normal scales but adjust tuning

    justMajor       = ScalePattern([0,2,4,5,7,9,11], name="justMajor", tuning=Tuning.just)
    justMinor       = ScalePattern([0,2,3,5,7,8,10], name="justMinor", tuning=Tuning.just)

    dorian          = ScalePattern([0,2,3,5,7,9,10], name="dorian")
    dorian2         = ScalePattern([0,1,3,5,6,8,9,11], name="dorian2") 
    diminished      = ScalePattern([0,1,3,4,6,7,9,10], name="diminished")

    egyptian        = ScalePattern([0,2,5,7,10], name="egyptian")
    yu              = ScalePattern([0,3,5,7,10], name="yu")
    zhi             = ScalePattern([0,2,5,7,9], name="zhi")
    phrygian        = ScalePattern([0,1,3,5,7,8,10], name="phrygian")
    prometheus      = ScalePattern([0,2,4,6,11], name="prometheus")
    indian          = ScalePattern([0,4,5,7,10], name="indian")

    locrian         = ScalePattern([0,1,3,5,6,8,10], name="locrian")
    locrianMajor    = ScalePattern([0,2,4,5,6,8,10], name="locrianMajor")

    lydian          = ScalePattern([0,2,4,6,7,9,11], name="lydian")
    lydianMinor     = ScalePattern([0,2,4,6,7,8,10], name="lydianMinor")

    custom          = ScalePattern([0,2,3,5,6,9,10], name="custom")

    hungarianMinor  = ScalePattern([ 0, 2, 3, 6, 7, 8, 11 ], name="hungarianMinor")
    romanianMinor   = ScalePattern([ 0, 2, 3, 6, 7, 9, 10 ], name="romanianMinor")
    chinese         = ScalePattern([ 0, 4, 6, 7, 11 ], name="chinese")

    wholeTone       = ScalePattern([ 0, 2, 4, 6, 8, 10 ], name="wholeTone")

    # Half-Whole Diminished Scale - halfWhole
    halfWhole       = ScalePattern([ 0, 1, 3, 4, 6, 7, 9, 10 ], name= "halfWhole")
    # Whole-Half Diminished Scale - wholeHalf
    wholeHalf       = ScalePattern([ 0, 2, 3, 5, 6, 8, 9, 11 ], name= "wholeHalf")
    
    ### Bebop Scales ###
    bebopMaj        = ScalePattern([ 0, 2, 4, 5, 7, 8, 9, 11 ], name="bebopMaj")
    bebopDorian     = ScalePattern([ 0, 2, 3, 4, 5, 9, 10 ], name="bebopMin") # aka Bebop Minor
    bebopDom        = ScalePattern([ 0, 2, 4, 5, 7, 9, 10, 11 ], name="bebopDom") # Bebop Dominant/Mixolydian
    bebopMelMin     = ScalePattern([ 0, 2, 3, 5, 7, 8, 9, 11 ], name="bebopMelMin") # Bebop Melodic Minor
    blues           = ScalePattern([ 0, 3, 5, 6, 7, 10 ], name="blues")

    ### Modes of the Melodic Minor Scale ###
    
    # First mode - Min/Maj chord
    minMaj         = ScalePattern([ 0, 2, 3, 5, 7, 9, 11 ], name= "minMaj")
    # Second mode - (x)susb9
    susb9          = ScalePattern([ 0, 1, 3, 5, 7, 9, 10 ], name= "susb9")
    # Third Mode - Lydian Augmented, (x)Maj7#5
    lydianAug      = ScalePattern([ 0, 2, 4, 6, 8, 9, 11 ], name= "lydianAug")
    # Fourth Mode - Lydian Dominant, (x)7#11
    lydianDom      = ScalePattern([ 0, 2, 4, 6, 7, 9, 10 ], name= "lydianDom")
    # Fifth Mode - seldom used, but it's IMinMaj/V
    melMin5th      = ScalePattern([ 0, 2, 4, 5, 7, 8, 10 ], name= "melMin5th")
    # Sixth Mode - half-diminished (aka Locrian #2), (x)half-diminished
    halfDim      = ScalePattern([ 0, 2, 3, 5, 6, 8, 10 ], name= "halfDim")
    # Seventh Mode - altered (diminished whole-tone), (x)7alt
    altered      = ScalePattern([ 0, 1, 3, 4, 6, 8, 10 ], name= "altered")
    


    freq            = FreqScalePattern()

    def __init__(self):

        self.default = _DefaultScale(self.major)

    def __setattr__(self, key, value):
        if key == "default" and key in vars(self):
            self.default.set(value)
        else:
            self.__dict__[key] = value
        return

    def __getitem__(self, key):
        return getattr(self, key)

    def get_scale(self, name):
        """ Returns a ScalePattern using a name """
        return self.library()[name]

    def library(self):
        """ Returns a dictionary with scale names to scale instances """
        lib = []
        for items in (self.__class__.__dict__.items(), self.__dict__.items()):
            lib.extend([(key, value) for key, value in items if isinstance(value, ScalePattern)])
        return dict(lib)

    def names(self):
        """ Returns a list of all the scale names """
        return sorted(self.library().keys())

    def scales(self):
        """ Returns a list of all the scales object """
        return sorted(self.library().values(), key=lambda scale: scale.name)

    def choose(self):
        """ Scale.choose() -> Returns a random scale object """
        return choice(self.scales())


Scale = __scale__()


# class Chord:
#     def __init__(self):
#         self.scale = Scale.default
#         self.root  = Root.default



from __future__ import absolute_import, division, print_function

from .tkimport import *
    
from .Format import *
from .AppFunctions import *

from ..Settings import AUTO_COMPLETE_BRACKETS

whitespace = [" ","\t","\n","\r","\f","\v"]

class BracketHandler:
    counter = 0
    def __init__(self, master):

        self.root = master
        
        self.text = master.text

        self.active = AUTO_COMPLETE_BRACKETS

        self.inbrackets = False

        self.separators = py_separators

        self.style = bracket_formatting

        self.left_brackets = dict(zip(left_b, right_b))
        self.right_brackets = dict(zip(right_b, left_b))
        self.right_brackets_only = list("]})")

        self.queue = []

        if self.active:

            for char in list(self.left_brackets.keys()) + list(self.right_brackets.keys()):

                self.text.bind(char, self.handle) # this binds bracket presses to the handle method

        self.visible_range = (0,0)

    def handle(self, event=None, insert=INSERT):

        self.visible_range = self.text.get_visible_range()

        ret = None

        line, column = index(self.text.index(insert))

        # 1. Type a left bracket

        next_char = self.text.get(index(line, column))

        if event.char in self.left_brackets:

            # A. Type a left bracket and an area is selected

            try:

                a = self.text.index(SEL_FIRST)
                b = self.text.index(SEL_LAST)

                self.text.insert(b, self.left_brackets[event.char])
                self.text.insert(a, event.char)

                # unselect

                self.text.tag_remove(SEL, "1.0", END)

                ret = "break"
                
            except:

                pass

            # B. If next character is a whitespace or a right bracket add a closing bracket as well. 

            if next_char in whitespace + self.right_brackets_only:

                # If the next char is quote mark, just move the cursor

                if event.char in ("'", '"'):

                    if event.char == next_char:

                        self.text.mark_set(insert, index(line, column + 1))

                        ret = "break"

                    else:

                        # Only add a closing quote mark if there are already an even no. of quotes in the line

                        text = self.text.get(index(line, 0), index(line, "end"))

                        n = text.count(event.char)

                        if n % 2 == 0:

                            self.text.insert(self.text.index(insert), event.char + self.left_brackets[event.char])

                            self.text.mark_set(insert, index(line, column + 1))

                            ret = "break"

                else:

                    self.text.insert(self.text.index(insert), event.char + self.left_brackets[event.char])

                    self.text.mark_set(insert, index(line, column + 1))

                    ret = "break"

            # Update line colours
        
            self.root.colour_line(line)

        # 2. Type right bracket
        elif event.char in self.right_brackets:

            # Go back and find an open bracket -> assume the first open bracket is related to *this* one

            self.text.insert("{}.{}".format(line,column), event.char) # Add bracket

            coords = self.find_starting_bracket(line, column, event.char)

            new_line, new_col = line, column + 1

            # Assume we are adding a new bracket

            adding_bracket = True

            # Get index of the end of the buffer

            end_line, end_col = index(self.text.index("{}.end".format(self.visible_range[1])))

            while (new_line, new_col) != (end_line, end_col):

                # If we find a closing bracket,  find it's pair

                next_char = self.text.get(index(new_line, new_col))

                if next_char == event.char:

                    coords_ = self.find_starting_bracket(new_line, new_col, event.char, offset=0)

                    # If there is a closing brackets

                    if coords_ is None:

                        adding_bracket = False

                        break

                    else:

                        adding_bracket = True

                if index(new_line, new_col) == self.text.index(index(new_line, "end")):
                    
                    new_line += 1
                    new_col   = 0

                else:

                    new_col += 1
                
            if not adding_bracket:

                loc = index(line, column+1)

                if self.text.get(loc) == event.char:

                    self.text.delete(loc)

            ret = "break"

            # Update line colours
            
            self.root.colour_line(line)

            # Highlight brackets (remove old highlighting first)

            self.text.tag_delete("tag_open_brackets")

            if coords is not None:

                row, col = coords

                self.text.tag_config("tag_open_brackets", **self.style)
                self.text.tag_add("tag_open_brackets", "{}.{}".format(row,col), "{}.{}".format(row,col+1))
                self.text.tag_add("tag_open_brackets", "{}.{}".format(line,column), "{}.{}".format(line,column+1))

        if ret is None:
            
            self.text.insert(INSERT, event.char)
            self.root.colour_line(line)

        # Store the text

        self.root.text_as_string = self.root.get_all()

        # Add separator to undo stack

        self.text.edit_separator()

        return "break"

    def delete(self, insert=INSERT):
        """ Deletes enclosed set of brackets if no contents and returns True, returns False
            all other times """
        if self.active:
            line, column = index(self.text.index(insert))
            next_char    = self.text.get(index(line, column))
            prev_char    = self.text.get(index(line, column-1))

            if prev_char in self.left_brackets:
                if next_char == self.left_brackets[prev_char] and column > 0:
                    self.text.delete(index(line, column-1), index(line, column+1))
                    return True
        return False

    def find_starting_bracket(self, line, column, bracket_style, offset = 0):
        """ Finds the opening bracket to the closing bracket at line, column co-ords.
            Returns None if not found. Only goes as far as the first line visible on screen"""
        
        line_length = column - 1
        used_br = offset

        start_row, end_row = self.visible_range # returns tuple

        for row in range(line, start_row - 1, -1):

            for col in range(line_length, -1, -1):

                # If the char is a left bracket and not used, break

                if self.text.get("{}.{}".format(row, col)) == self.right_brackets[bracket_style]:

                    if used_br == 0:

                        return row, col

                    else:

                        used_br -= 1

                elif self.text.get("{}.{}".format(row, col)) == bracket_style:

                    used_br += 1

            line_length = int(self.text.index("{}.end".format(row-1)).split(".")[1])

        else:

            return None



from __future__ import absolute_import, division, print_function

from .Patterns import Pattern, PGroup

# Constants and constant values

class const:
    """ A number value that cannot be changed """
    def __init__(self, value):

        if isinstance(value, (list, Pattern)):
            self.__class__ = Pattern
            self.__init__([const(x) for x in value])
        elif isinstance(value, tuple):
            self.__class__ = PGroup
            self.__init__([const(x) for x in value])
        elif isinstance(value, PGroup):
            self.__class__ = value.__class__
            self.__init__([const(x) for x in value])
        else:
            self.value = value

    def __int__(self):
        return int(self.value)
    def __float__(self):
        return float(self.value)
    def __repr__(self):
        return str(self.value)
    def __add__(self, other):
        return self.value
    def __radd__(self, other):
        return self.value
    def __sub__(self, other):
        return self.value
    def __rsub__(self, other):
        return self.value
    def __mul__(self, other):
        return self.value
    def __rmul__(self, other):
        return self.value
    def __div__(self, other):
        return self.value
    def __rdiv__(self, other):
        return self.value
    def __gt__(self, other):
        return self.value > other
    def __ge__(self, other):
        return  self.value >= other
    def __lt__(self, other):
        return self.value < other
    def __le__(self, other):
        return self.value <= other
    def __eq__(self, other):
        return self.value == other
    def __ne__(self, other):
        return self.value != other

class _inf(const):
    def __repr__(self):
        return "inf"
    def __add__(self, other):
        return _inf(self.value + other)
    def __radd__(self, other):
        return _inf(other + self.value)
    def __sub__(self, other):
        return _inf(self.value - other)
    def __rsub__(self, other):
        return _inf(other - self.value)
    def __mul__(self, other):
        return _inf(self.value * other)
    def __rmul__(self, other):
        return _inf(other * self.value)
    def __div__(self, other):
        return _inf(self.value / other)
    def __rdiv__(self, other):
        return _inf(other / self.value)
    def __truediv__(self, other):
        return _inf(self.value / other)
    def __rtruediv__(self, other):
        return _inf(other / self.value)
    def __eq__(self, other):
        return isinstance(other, _inf)
    def __gt__(self, other):
        return not isinstance(other, _inf)
    def __ge__(self, other):
        return True
    def __lt__(self, other):
        return False
    def __le__(self, other):
        return isinstance(other, _inf)

inf = _inf(0)

class NoneConst(const):
    def __init__(self):
        self.value = None


# startup.py
# ----------
# This file is loaded on startup of FoxDot. You can edit it to include any code
# you want to run at the beginning of a session. To use an alternate startup
# file use the --startup flag followed by a valid path. You can also run FoxDot
# with a --no-startup flag to not load this file.


# Tutorial 0: Introduction

################
# Executing code

# To execute code in FoxDot, make sure your text cursor is in the 'block' of code
# (sections of text not separated by blank lines) and press Ctrl+Return

# To execute just a single line, even in a block, press Alt+Return

# Try it now, move the cursor to the line of code below and press Ctrl+Return

print("Hello World")

##############
# Help

# If you're ever stuck, or want to know more about a function or class
# just type help followed by the name of that Python object in brackets:

help(object)



# -*- coding: latin-1 -*-
import math
import sys

def generateVoiceSpecification(lyrics,tempo,inputMusicXMLPath,outputMusicXMLPath):
	with open(inputMusicXMLPath, 'r') as c:
		content = [x.strip() for x in c.readlines()]
		with open(outputMusicXMLPath,'w') as f:
			f.write(addVoiceTags(tempo,lyrics,content))

def addVoiceTags(tempo, text, content):
	print("Text:\n" + str(text))
	output = ""
	tempo_xml = '<direction>\n<sound tempo="{}"/>\n</direction>'.format(tempo)
	lyrics_xml = '<voice>1</voice>\n<lyric>\n<text>{}</text>\n</lyric>'

	i = 0
	ignoreThisNote = False
	for line in content:
		if "<rest/>" in line:
			ignoreThisNote = True
		if "</note" in line:
			if not ignoreThisNote:
				output += lyrics_xml.format(text[i%len(text)])
				i+=1
			else:
				ignoreThisNote = False

		output += line

		if tempo != -1 and "<measure" in line:
			output += tempo_xml

	return output



# Tutorial 14: Player Attributes Reference
# --- TODO: this needs updating


# To see all attributes:
print(Player.get_attributes())

# You can see what effects are available by evaluating
print(FxList)

# Let's use the high pass filter for an example. You can see it's described
# like so:
# "<Fx 'highPassFilter' -- args: hpr, hpf>"

# Each effect has a "master" argument and then child arguments. Here the
# master argument is "hpf" (short for high pass filter) and the child argument
# is "hpr" (short for high pass resonance). The effect is only added when the
# master argument is non-zero:
d1 >> dirt([0,4,2,1], dur=1/2, hpf=4000)

# This sets the high pass filter to 4000 Hz so only frequences in the audio
# signal *above* that are actually heard. Let's change the resonance value. It's
# default value is 1, so let's make it smaller
d1 >> dirt([0,4,2,1], dur=1/2, hpf=4000, hpr=0.3)


# Notice a difference? We can use patterns / vars in our effects to make them
# change over time:
d1 >> dirt([0,4,2,1], dur=1/2, hpf=linvar([0,4000],8), hpr=P[1,1,0.3].stretch(8))




####################
# Reference
####################



####################
# amp - Amplitude (defaults to 1)
# Sets the volume of the note/pattern

d1 >> play("*", dur=1/2, amp=1)

# Half Volume
d1 >> play("*", dur=1/2, amp=.5)

# Creating a pattern with amp
d1 >> play("*", dur=1/2, amp=[1,0,1,1,0])



####################
# amplify - Changes amp, by multiplying agasint the existing value (instead of overwritting)

# Creating a pattern with amp
d1 >> play("*", dur=1/2, amp=[1,0,1,1,0])
d1 >> play("*", dur=1/2, amplify=[.5,1,0])

# Set up a "drop" in the music (Plays at full volume for 28, then 0 for 4)
p1 >> blip([0,1,2,3], amplify=var([1,0],[28,4]))



####################
# bend



####################
# benddelay - See bend



####################
# bits
# The bit depth, in number of bits, that the signal is reduced to;
# this is a value between 1 and 24 where other values are ignored.
# Use crush to set the amount of reduction to the bitrate (defaults to 8)



####################
# bitcrush - See bits



####################
# blur



####################
# bpf - Band Pass Filter



####################
# bpnoise - See bpf



####################
# bpr - See bpf



####################
# bpm



####################
# buf



####################
# channel



####################
# chop
# 'Chops' the signal into chunks using a low frequency pulse wave over the sustain of a note.



####################
# coarse



####################
# comb delay - See echo



####################
# crush



####################
# cut
# Cuts a duration
p1 >> pluck(P[:8], dur=1/2, cut=1/8)
p1 >> pluck(P[:8], dur=1/2, cut=1/4)
p1 >> pluck(P[:8], dur=1/2, cut=1/2)



####################
# cutoff



####################
# decay - See echo



####################
# degree - The degree of the note, or pitch, can be specified by keyword (also the first positional)
p1 >> blip(degree=[0,1,2,3])

# Which is the same as:
p1 >> blip([0,1,2,3])

# Only plays the "root" note of the chord
b1 >> bass(p1.degree[0])



####################
# delay - A duration of time to wait before sending the information to SuperCollider (defaults to 0)

# Delays every 3 note by .1
p1 >> blip([0,1,2,3], delay=[0,0,0.1])

# Delays every 3 note by .5
p1 >> blip([0,1,2,3], delay=[0,0,0.5])

# Plays the note once for each different delays
p1 >> blip([0,1,2,3], delay=(0,0.1))

p1 >> blip([0,1,2,3], delay=(0,0.25))

p1 >> blip([0,1,2,3], delay=(0,.1,.2,.3))



####################
# dist



####################
# dur - Durations (defaults to 1 and 1/2 for the Sample Player)



####################
# echo
# Title keyword: echo, Attribute keyword(s): decay
# Sets the decay time for any echo effect in beats, works best on Sample Player (defaults to 0)
# Multiplied against the sustain value
d1 >> play("x-o-", echo=0.1)

d1 >> play("x-o-", echo=0.5)

p1 >> pluck(P[:8], echo=.25)

p1 >> pluck(P[:8], echo=.5)

p1 >> pluck(P[:8], echo=.5, decay=.5)



####################
# env



####################
# fmod



####################
# formant



####################
# freq



####################
# hpf - High Pass Filter
# Filters out all the frequencies below given value, removing lower freqencies

# 4000 hertz
p1 >> pluck(P[:8], dur=1/2, hpf=4000)

# HPF is 0 for 4 beats, then 4000 for 4 beats
p1 >> pluck(P[:8], dur=1/2, hpf=var([0,4000],[4,4]))

# Linear change on hpf from 0 take 4 beats to get to 4000, 4 beats back to 0
p1 >> pluck(P[:8], dur=1/2, hpf=linvar([0,4000],[4,4]))

# Linear change on hpf from 0 take 8 beats to get to 4000, then reset back to 0
p1 >> pluck(P[:8], dur=1/2, hpf=linvar([0,4000],[8,0]))

# With resonance change (default is 1)
p1 >> pluck(P[:8], dur=1/2, hpf=linvar([0,4000],[8,0]), hpr=.5)

# With resonance change as a linvar
p1 >> pluck(P[:8], dur=1/2, hpf=linvar([0,4000],[8,0]), hpr=linvar([0.1,1],12))



####################
# hpr - See hpf



####################
# lpf - Low Pass Filter
# Filters out all the frequencies above given value, removing higher freqencies

# 4000 hertz
p1 >> pluck(P[:8], dur=1/2, lpf=400)

# With resonance change as a linvar
p1 >> pluck(P[:8], dur=1/2, lpf=linvar([500,4000],[8,0]), lpr=linvar([0.1,1],12))



####################
# lpr - See lpf



####################
# midinote



####################
# pan
# Panning, where -1 is far left, 1 is far right (defaults to 0)



####################
# pitch - See degree



####################
# pshift



####################
# oct



####################
# rate
# Variable keyword used for misc. changes to a signal. E.g. Playback rate of the Sample Player (defaults to 1)



####################
# room
# Title keyword: room, Attribute keyword(s): mix

# The room argument specifies the size of the room
d1 >> play("x-o-", room=0.5)

# Mix is the dry/wet mix of reverb or how much the reverb is mixed with the source.  1 is all reverb, 0 is no reverb at all. (Default 0.1)
d1 >> play("x-o-", room=0.5, mix=.5)



####################
# Reveb
# See Room



####################
# sample
# Special keyword for Sample Players; selects another audio file from the bank of samples for a sample character.



####################
# scale



####################
# shape



####################
# slide - Slide To
# Slides' the frequency value of a signal to freq * (slide+1) over the duration of a note (defaults to 0)

p1 >> pluck(P[:8], dur=1/2, slide=1)

p1 >> pluck(P[:8], dur=1/2, slide=12)

p1 >> pluck(P[:8], dur=1/2, slide=var([0,-1],[12,4]))



####################
# slidedelay



####################
# slidefrom



####################
# slider



####################
# spread



####################
# spin



####################
# striate



####################
# stutter



####################
# sus - Sustain (defaults to `dur`)



####################
# swell



####################
# vib - Vibrato
# Vibrato - Title keyword: vib, Attribute keyword(s): Vibrato (defaults to 0)

p1 >> pluck(P[:8], dur=1/2, vib=12)

# With child attribute, vibdepth (default 0.2)
p1 >> pluck(P[:8], dur=1/2, vib=12, vibdepth=0.5)



####################
# vibdepth - See vib



# Tutorial 1: Playing Notes

# In FoxDot, all two-character variable names are reserved for player objects, such as 'p1'
# Creating a Player Object with no arguments will play a single note on middle C, by default, repeatedly until stopped.
# Use >> to give one of these to a player object like so:

p1 >> pluck()

# To stop an individual player object, simply execute

p1.stop()

# Besides the 2-character variables that are pre-reserved, you can create your
# own with your own names

foo = Player()
foo >> pluck()

# The >> in Python is usually reserved for a type of operation, like + or -, but it is not the case in FoxDot.
# If a user re-executes the code, FoxDot will update p1 instead of creating a PlayerObject,
# which means you can make changes to your music using just one line of code.

# If you now give your player object some arguments, you can change the notes being played back.
# The first argument should be the degree of the note to be played
# (default is the lowest note of octave 5 of the major scale) and does not need to be specified by name.

# Python, like most programming languages, using zero-indexing when accessing values in an array,
# which means that 0 refers to the first note of the scale.
# Give your player object instructions to make music with their Synth.
# The first argument is the note of the scale to play. The following code
# plays the first three notes of the default scale (major) on repeat.

# For a single note
p1 >> pluck(0)

# Or a list of notes
p1 >> pluck([0,1,2])

# But you’ll need to specify whatever else you want to change...

# Such as note durations, or the length of each note
p1 >> pluck([0,0,0], dur=[1,2,3])

# Or amplitude, the "volume" of each note
p1 >> pluck([0,0,0], amp=[1,2,3])

# If the second list, the amp in this example, is too long, then the first list (the degree) just loops, and are matched with the remaining elements from the second list (the amplitude).
p1 >> pluck([0,2,4], amp=[1,2,3,1,5])

# More generally, all the lists are traversed regardless of their length.
p1 >> pluck([0,2,4], dur=[1,2], amp=[1,2,3,1,5])

# Arguments can be integers, floating points, fractions, lists,
# tuples, or a mix

p1 >> pluck([0,0,0], dur=2)

p1 >> pluck([0,0,0], dur=1.743)

p1 >> pluck([0,0,0], dur=[0.25,0.5,0.75])

p1 >> pluck([0,0,0], dur=[1/4,1/2,3/4])

p1 >> pluck([0,0,0], dur=[1/4,0.25,3])

# Lists of values are iterated over as the Player plays notes
# The following duration equates to:  1,2,3,1,4,3
# If you don't understand this yet, don't worry, more about patterns in the pattern tutorial
p1 >> pluck([0,0,0], dur=[1,[2,4],3])

# Values in tuples are used simultaneously i.e. p1 will play 3 individual notes, then a chord of 3 together at the same time.
p1 >> pluck([0,2,4,(0,2,4)])

# You can also assign values to the attributes of player objects directly
p1.oct = 5

# To see all the names of player attributes, just execute
print(Player.get_attributes())

# More about those later in the player attributes tutorial

# You could store several player instances and assign them at different times
proxy_1 = pads([0,1,2,3], dur=1/2)
proxy_2 = pads([4,5,6,7], dur=1)

p1 >> proxy_1 # Assign the first to p1

p1 >> proxy_2 # This replaces the instructions being followed by p1

# To play multiple sequences at once, just do the same things with another
# Player object:

p1 >> pluck([0, 2, 3, 4], dur=1/2)

p2 >> pads([(0, 2, 4), (3, 5, 7)], dur=8)

# Play only this player, muting others
p1.solo() # default value is 1 (solo on)

# And turn the solo off
p1.solo(0)

# Stop (not just mute) the other players
p1.only()

# Use Ctrl+. to clear everything for the scheduling clock or run
Clock.clear()



from __future__ import absolute_import, print_function

from .MidiFactory import createMidi
import os
from math import ceil

def extendScale(scale,times):
    scaleAux = scale[:]
    for i in range(1,times):
        for note in scaleAux:
            scale.append(note+12*i)
    return scale

def matchListsSize(rhythm,melody):
    n_rhythm =len(rhythm)
    n_melody = len(melody)
    if n_rhythm < n_melody:
        rhythm = (rhythm*int(ceil(n_melody/float(n_rhythm))))[:n_melody]
    elif n_rhythm != n_melody:
        melody = (melody*int(ceil(n_rhythm/float(n_melody))))[:n_rhythm]

    return rhythm,melody

def toList(rhythm, melody, scale):

    scale = extendScale(scale,2)
    scaleSize = len(scale)

    volume = 100
    baseNote = 60

    rhythm, melody = matchListsSize(rhythm,melody)

    composition = []
    time = 0
    for i in range(len(rhythm)):
        dur = rhythm[i]
        note = scale[melody[i]%scaleSize] + baseNote
        composition.append((note,dur,volume,time))
        time += dur

    return composition

def compose(notes,durations,scale, new_midi_path, new_musicxml_path):

    print(notes)
    print(durations)
    print(scale)

    composition = toList(durations,notes,scale)
    print(composition)
    createMidi(new_midi_path, composition)
    os.system("musescore "+ new_midi_path +" -o " + new_musicxml_path)


""" Module for converting handling MIDI in/out and functions relating to MIDI pitch calculation. """

from __future__ import absolute_import, division, print_function

try:
    import rtmidi
    from rtmidi import midiconstants
    TIMING_CLOCK          = midiconstants.TIMING_CLOCK
    SONG_POSITION_POINTER = midiconstants.SONG_POSITION_POINTER
    SONG_START            = midiconstants.SONG_START
    SONG_STOP             = midiconstants.SONG_STOP 
except ImportError as _err:
    pass

from .Patterns import asStream
from .Scale    import ScalePattern
from .TimeVar  import TimeVar
from .SCLang import SynthDefProxy

import time

class MidiInputHandler(object):

    """Midi Handler CallBack Function"""

    def __init__(self, midi_ctrl):

        self.midi_ctrl = midi_ctrl
        self.bpm_group = []
        self.played = False

    def __call__(self, event, data=None):

        datatype, delta = event

        self.midi_ctrl.delta += delta
        
        if TIMING_CLOCK in datatype and not self.played:

            self.midi_ctrl.pulse += 1
            

            if self.midi_ctrl.pulse == self.midi_ctrl.ppqn:

                t_master = 60.0
                
                self.midi_ctrl.bpm = round(60.0 / self.midi_ctrl.delta,0)

                self.midi_ctrl.pulse = 0
                self.midi_ctrl.delta = 0.0

                #print("BPM : " + repr(self.midi_ctrl.bpm))
            
class MidiIn:
    metro = None
    def __init__(self, port_id=0):
        """ Class for listening for MIDI clock messages
            from a midi device """
        try:

            self.device = rtmidi.MidiIn()

        except NameError:

            raise ImportError(_err)

        self.available_ports = self.device.get_ports()

        if not self.available_ports:

            raise MIDIDeviceNotFound

        else:

            print("MidiIn: Connecting to " + self.available_ports[port_id])

        self.device.open_port(port_id)
        self.device.ignore_types(timing=False)


        self.pulse = 0
        self.delta = 0.0
        self.bpm   = 120.0
        self.ppqn  = 24
        self.beat  = 0

        self.device.set_callback(MidiInputHandler(self))

    @classmethod
    def set_clock(cls, tempo_clock):
        cls.metro = tempo_clock
        return

    def close(self):
        """ Closes the active port """
        self.device.close_port()
        return


class MidiOut(SynthDefProxy):
    """ SynthDef proxy for sending midi message via supercollider """
    def __init__(self, degree=0, **kwargs):
        SynthDefProxy.__init__(self, self.__class__.__name__, degree, kwargs)

midi = MidiOut # experimental alias

# Midi information exceptions

class MIDIDeviceNotFound(Exception):
    def __str__(self):
        return self.__class__.__name__ + " Error"

class rtMidiNotFound(Exception):
    def __str__(self):
        return self.__class__.__name__ + ": Module 'rtmidi' not found"


if __name__ == "__main__":

    a = MidiIn()



from __future__ import absolute_import, division, print_function

from .Patterns import *
from .TimeVar import TimeVar
from .Utils import recursive_any, get_inverse_op
from functools import partial

def convert_to_pattern(value):
    if isinstance(value,  list):
        value = Pattern(value)
    elif isinstance(value, tuple):
        value = PGroup(value)
    elif not isinstance(value, metaPattern):
        value = Pattern(value)
    return value

def convert_pattern_args(func):
    def new_method(self, value):
        if isinstance(value, (list, tuple)):
            value = convert_to_pattern(value)
        # if isinstance(value, (metaPattern, GeneratorPattern)):
        if isinstance(value, (Pattern, GeneratorPattern)):
            other_op = get_inverse_op(func.__name__)
            return getattr(value, other_op).__call__(self)
        return func(self, value)
    return new_method


class NumberKey(object):
    """ An object that acts like a number but may have dependencies when returing its
        value. These are used when returing Player object attribute values accessed via
        `getattr` e.g. `p1.dur` or `getattr(p1, "dur")`. The parent attribute is the Player
        that contains the key or the referenced key in the case that the value has been
        maniupulated. e.g. if `p1` is using `p2.degree` then `p1.degree.parent == p1` and 
        `p1.degree.value.parent == p2`.
    """

    def __init__(self, value=0, function=None):
        # the number to store/update
        self.value     = value
        self.calculate = function if function is not None else lambda x: x
        # reference to another number key that this is linked to
        # self.other = reference
        # self.parent = self.value if isinstance(self.value, NumberKey) else None
        # This is the Player object whose attribute this is fetching / chained to
        # self.parent = self.other.parent if isinstance(self.other, NumberKey) else None
        # self.

    def parent(self):
        return self.value if isinstance(self.value, NumberKey) else None

    def get_root(self):
        child = self
        while True:
            parent = child.parent() 
            if parent is None:
                break
            child = parent
        return child

    def path_to_root(self):
        child = self
        while True:
            parent = child.parent()
            if parent is None:
                break
            child = parent
            yield child


    def has_circular_reference(self):
        return self in self.path_to_root()

    def is_root(self):
        return self.parent() is None

    # Storing mathematical operations
    
    @convert_pattern_args
    def __add__(self, other):
        function = lambda value: value + other
        return self.transform(function)

    @convert_pattern_args
    def __radd__(self, other):
        function = lambda value: other + value
        return self.transform(function)

    @convert_pattern_args
    def __sub__(self, other):
        function = lambda value: value - other
        return self.transform(function)

    @convert_pattern_args
    def __rsub__(self, other):
        function = lambda value: other - value
        return self.transform(function)
    
    @convert_pattern_args
    def __mul__(self, other):
        function = lambda value: value * other
        return self.transform(function)

    @convert_pattern_args
    def __rmul__(self, other):
        function = lambda value: other * value
        return self.transform(function)

    @convert_pattern_args
    def __truediv__(self, other):
        function = lambda value: value / other
        return self.transform(function)

    @convert_pattern_args
    def __rtruediv__(self, other):
        function = lambda value: other / value
        return self.transform(function)

    @convert_pattern_args
    def __floordiv__(self, other):
        function = lambda value: value // other
        return self.transform(function)

    @convert_pattern_args
    def __rfloordiv__(self, other):
        function = lambda value: other // value
        return self.transform(function)
    
    @convert_pattern_args
    def __mod__(self, other):
        function = lambda value: value % other
        return self.transform(function)

    @convert_pattern_args
    def __rmod__(self, other):
        function = lambda value: other % value
        return self.transform(function)
    
    @convert_pattern_args
    def __pow__(self, other):
        """ If operating with a pattern, return a pattern of values """
        function = lambda value: value ** other
        return self.transform(function)
    
    @convert_pattern_args
    def __rpow__(self, other):
        """ If operating with a pattern, return a pattern of values """
        function = lambda value: value ** other
        return self.transform(function)
    
    @convert_pattern_args
    def __xor__(self, other):
        """ If operating with a pattern, return a pattern of values """
        function = lambda value: value ** other
        return self.transform(function)
    
    @convert_pattern_args
    def __rxor__(self, other):
        """ If operating with a pattern, return a pattern of values """
        function = lambda value: other ** value
        return self.transform(function)

    @convert_pattern_args
    def __eq__(self, other):
        function = lambda value: value == other
        return self.transform(function)
    
    @convert_pattern_args
    def __ne__(self, other):
        function = lambda value: (value != other)
        return self.transform(function)

    @convert_pattern_args
    def __gt__(self, other):
        function = lambda value: (value > other)
        return self.transform(function)

    @convert_pattern_args
    def __ge__(self, other):
        function = lambda value: (value >= other)
        return self.transform(function)

    @convert_pattern_args
    def __lt__(self, other):
        function = lambda value: (value < other)
        return self.transform(function)

    @convert_pattern_args
    def __le__(self, other):
        function = lambda value: (value <= other)
        return self.transform(function)

    def __abs__(self):
        return self.transform(abs)

    def __getitem__(self, key):
        def function(value):
            try:
                return value[key]
            except TypeError:
                return value
        return self.spawn_child(function)

    def index(self, sequence):
        """ Returns a Player Key that returns the element from sequence indexed using int(self) """
        new = self.child(sequence)
        def getitem(b, a):
            try:
                return b[a]
            except TypeError:
                return b
        new.calculate = getitem
        return new

    def semitones(self):
        """ Converts the current value into the semitone value using the parent's scale """
        new = self.child(0)
        new.calculate = lambda a, b: self.parent.scale.semitones(b)
        return new

    def simple_map(self, mapping):
        """ Creates a new Player key that maps the values in the dictionary (mapping)
            to new values. Example use case:

            ```
            d1 >> play("x-o-", sample=d1.degree.simple_map( { "-" : -1, "o" : var([0,2]) }))
            ```

        """
        assert isinstance(mapping, dict)
        data = [ ((self == key) * value) for key, value in mapping.items() ]
        new_key = data[0]
        for i in data[1:]:
            new_key = new_key + i
        return new_key

    def map(self, mapping, default=0):     
        
        # input functions
        functions = []
    
        # Convert default output to function
        if callable(default):
            default_func = default
        else:
            default_func = partial(lambda x, y: x, default)
        
        # Convert input values to functions
        for key, value in mapping.items():
            if callable(key):
                test_func = force_pattern_args(key)
            else:
                test_func = partial(lambda x, y: x == y, key)
            
            if callable(value):
                result_func = force_pattern_args(value)
            else:
                result_func = partial(lambda x, y: x, value)
                
            functions.append((test_func, result_func))
            
        # Define mapping function to test input functions
        def mapping_function(value):
            # For PGroups
            if isinstance(value, PGroup):
                return PGroup([mapping_function(item) for item in value])

            # For other values
            for func, result in functions:
                if bool(func(value)) is True:
                    return result(value)
            return default_func(value)

        return self.spawn_child(mapping_function)

    def get_min(self):
        new = self.child(0)
        def f(a, b):
            try:
                return min(b)
            except TypeError:
                return b
        new.calculate = f
        return new

    def get_max(self):
        new = self.child(0)
        def f(a, b):
            try:
                return max(b)
            except TypeError:
                return b
        new.calculate = f
        return new

    def transform(self, func):
        """ Returns a child Player Key based on the func. If the value
            returned is a PGroup, that is also transformed by the function """
        # def new_func(item):
        #     if isinstance(item, (PGroup, Pattern)):
        #         return item.transform(func)
        #     else:
        #         return func(item)
        def new_func(item):
            try:
                return func(item)
            except AttributeError as e:
                error = e
            try:
                return item.transform(func)
            except AttributeError:
                # Raise original error for more information
                raise error

        return self.spawn_child(new_func)

    def accompany(self, rel=[0,2,4]):
        """ Returns a PlayerKey whose function returns an accompanying note """
        return self.transform(Accompany(rel=rel))

    def versus(self, rule=lambda x, y: x > y):
        """ p1 >> pads([0, 1, 2, 3])
            p2 >> pluck([4, 5, 0]).versus(p1, rule)
        """
        
        # 1. Sets this source player key amplify to be "off" when the rule is satisfied

        # 2. Returns a new PlayKey

        return 
    
    # Values
    
    def __nonzero__(self):
        return self.__bool__()
    def __bool__(self):
        return bool(self.now())
    def __int__(self):
        return int(self.now())
    def __float__(self):
        return float(self.now())
    def __str__(self):
        return str(self.now())
    def __repr__(self):
        return repr(self.now())
    def __len__(self):
        return len(self.now())
    
    def __iter__(self):
        try:
            for item in self.now():
                yield item
        except TypeError:
            yield self.now()

    # def child(self, other):
    #     return NumberKey(self.value, other)

    def spawn_child(self, function):
        return self.__class__(self, function)
    
    def now(self, other=None):
        """ Returns the current value in the Key by calling the parent """

        value = self.value.now() if hasattr(self.value, "now") else self.value

        return self.calculate(value)

class PlayerKey(NumberKey):
    # def __init__(self, value=None, reference=None, parent=None, attr=None):
    def __init__(self, value, function=None, player=None, attr=None):

        NumberKey.__init__(self, value, function)

        if player is None and isinstance(self.value, PlayerKey):

            self.attr   = self.value.attr
            self.player = self.value.player

        else:

            self.attr    = attr
            self.player = player

        # self.pattern = asStream(self.parent.attr[self.key]) if self.parent is not None else asStream([]) #
        # self.pattern = NumberKey().transform(lambda x: self.player.attr[self.attr])

        self.last_updated = 0

    def cmp(self, player, attr):
        return player == self.player and attr == self.attr

    def get_player_attribute(self):
        return self.player.attr[self.attr]

    def name(self):
        return "{}.{}".format(self.player.id, self.attr)

    def set(self, value, time):
        self.value = value
        self.last_updated = time
        return
    
    def update(self, value, time):
        """ Updates the contents of the PlayerKey *if* the time value is different to self.last_updated.
            If they are the same, the the contents become a PGroup of the two values """
        if not equal_values(value, self.value):
        #if value != self.value:
            if time == self.last_updated:
                try:
                    self.value.append(value)
                except AttributeError:
                    self.value = PGroup(self.value, value)
            else:
                self.value = value
        self.last_updated = time
        return

    # Could be removed
    def update_pattern(self):
        # try:
        #     self.pattern[:] = asStream(self.parent.attr[self.key])
        # except TypeError:
        #     self.pattern = asStream(self.parent.attr[self.key])
        return


class Accompany:
    """ Like PlayerKey except it returns """
    this_last_value = 0
    keys_last_value = None

    def __init__(self, rel=[0,2,4]):

        # self.frequency  = freq
        self.scale_size = 7
        self.relations  = list(rel)

    def __call__(self, playerkey):
        """ Acts as a function in Player Key """
        # Only change value if the player key has changed - maybe set a frequency?
        if self.keys_last_value != playerkey:
            self.this_last_value = self.find_new_value(playerkey)
            self.keys_last_value = playerkey
        return self.this_last_value

    def find_new_value(self, playerkey):

        # Which value is the closest to this_last_value

        values = [(playerkey + x) % 7 for x in self.relations] + [(playerkey % 7) + (x - self.scale_size) for x in self.relations]

        nearby = [abs(self.this_last_value - value) for value in values]
        
        indices = [nearby.index(val) for val in sorted(nearby)]

        r = random.random()

        if r <= 0.65:

            i = 0

        elif r <= 0.9:

            i = 1

        else:

            i = 2

        index = indices[i % len(indices)]

        return values[index]

class Versus(Accompany):
    def __init__(self):
        pass
    def find_new_value(self, playerkey):
        return 


# Give pattern objects a reference to the PlayerKey type

Pattern.PlayerKey = PlayerKey


"""
Sequences.py
------------
All patterns inherit from Base.Pattern. There are two types of pattern:

1. Container types
    * Similar to lists but with different mathematical operators
2. Generator types
    * Similar to generators but can be indexed (returns values based on functions)
"""

from __future__ import absolute_import, division, print_function

import random
import math

from .Main import *
from .PGroups import *
from .Operations import *
from .Generators import PRand

MAX_SIZE = 2048

#==============================#
#         1. P[] & P()         #
#==============================#

class __pattern__(object):
    ''' Used to define lists as patterns:

        `P[1,2,3]` is equivalent to `Pattern([1,2,3])` and
        `P(1,2,3)` is equivalent to `Pattern((1,2,3))` and
        `P+(1,2,3)` is equivalient to `Pattern((1,2,3))`.

        Ranges can be created using slicing, e.g. `P[1:6:2]` will generate the range
        1 to 6 in steps of 2, thus creating the Pattern `[1, 3, 5]`. Slices can be
        combined with other values in a Pattern such that `P[0,2,1:10]` will return
        the Pattern `P[0, 2, 1, 2, 3, 4, 5, 6, 7, 8, 9]`
    '''
    def __getitem__(self, args):
        if isinstance(args, Pattern):
            return args
        elif isinstance(args, (Pattern.TimeVar, Pattern.PlayerKey)):
            data = [args]
        elif hasattr(args, '__iter__') and not isinstance(args, (str, bytes, GeneratorPattern, PGroup)):
            data = []
            for item in args:
                if type(item) is slice:
                    data.extend(sliceToRange(item))
                else:
                    data.append(item)
        elif type(args) is slice:
            data = sliceToRange(args)
        else:
            data = args
        return Pattern(data)

    def __call__(self, *args):
        return PGroup(args if len(args) > 1 else args[0])

    def __mul__(self, other):
        """ P*[0,1,2] returns PRand([0,1,2])
            P*(0,1,2) returns PGroupStar(0,1,2)
        """
        if isinstance(other, (list, Pattern)):
            return PRand(list(other))
        else:
            return PGroupStar(other)

    def __pow__(self, other):
        """ P**(x1, x2,...,xn) - Returns scrambled version """
        return PGroupPow(other)

    def __xor__(self, other):
        """ P^(x1, x2,..., dur) -  Returns a PGroup that delays each value by dur * n """
        return PGroupXor(other)

    def __or__(self, other):
        """ P|("x", 2) """
        return PGroupOr(other)

    def __add__(self, other):
        return PGroupPlus(other)

    def __radd__(self, other):
        return self + other

    def __truediv__(self, other):
        return PGroupDiv(other)

    def __mod__(self, other):
        if isinstance(other, list):
            return Pattern().fromString(other[0], flat=True)
        return PGroupMod(other)

    def __and__(self, other):
        return PGroupAnd(other)

    def __invert__(self):
        return __reverse_pattern__()

class __reverse_pattern__(__pattern__):
    def __getattr__(self, name):
        return ~object.__getattr__(self, name)


# This is a pattern creator
P = __pattern__()

#================================#
#      2. Pattern Functions      #
#================================#

#: Pattern functions that take patterns as arguments

def PShuf(seq):
    ''' PShuf(seq) -> Returns a shuffled version of seq'''
    return Pattern(seq).shuffle()

def PAlt(pat1, pat2, *patN):
    ''' Returns a Pattern generated by alternating the values in the given sequences '''
    data = []
    item = [asStream(p) for p in [pat1, pat2] + list(patN)]
    size = LCM(*[len(i) for i in item])
    for n in range(size):
        for i in item:
            data.append(modi(i,n))
    return Pattern(data)

def PStretch(seq, size):
    ''' Returns 'seq' as a Pattern and looped until its length is 'size'
        e.g. `PStretch([0,1,2], 5)` returns `P[0, 1, 2, 0, 1]` '''
    return Pattern(seq).stretch(size)

def PPairs(seq, func=lambda n: 8-n):
    """ Laces a sequence with a second sequence obtained
        by performing a function on the original. By default this is
        `lambda n: 8 - n`. """
    i = 0
    data = []
    for item in seq:
        data.append(item)
        data.append(func(item))
        i += 1
        if i >= MAX_SIZE:
            break
    return Pattern(data)

def PZip(pat1, pat2, *patN):
    ''' Creates a Pattern that 'zips' together multiple patterns. `PZip([0,1,2], [3,4])`
        will create the Pattern `P[(0, 3), (1, 4), (2, 3), (0, 4), (1, 3), (2, 4)]` '''
    l, p = [], []
    for pat in [pat1, pat2] + list(patN):
        p.append(P[pat])
        l.append(len(p[-1]))
    length = LCM(*l)
    return Pattern([tuple(pat[i] for pat in p) for i in range(length)])


def PZip2(pat1, pat2, rule=lambda a, b: True):
    ''' Like `PZip` but only uses two Patterns. Zips together values if they satisfy the rule. '''
    length = LCM(len(pat1), len(pat2))
    data = []
    i = 0
    while i < length:
        a, b = modi(pat1,i), modi(pat2,i)
        if rule(a, b):
            data.append((a,b))
        i += 1
    return Pattern(data)

@loop_pattern_func
def PStutter(x, n=2):
    """ PStutter(seq, n) -> Creates a pattern such that each item in the array is repeated n times (n can be a pattern) """
    return Pattern([x for i in range(n)])

@loop_pattern_func
def PSq(a=1, b=2, c=3):
    ''' Returns a Pattern of square numbers in the range a to a+c '''
    return Pattern([x**b for x in range(a,a+c)])

@loop_pattern_func
def P10(n):
    ''' Returns an n-length Pattern of a randomly generated series of 1's and 0's '''
    return Pattern([random.choice((0,1)) for i in range(int(n))])

@loop_pattern_func
def PStep(n, value, default=0):
    ''' Returns a Pattern that every n-term is 'value' otherwise 'default' '''
    return Pattern([default] * (n-1) + [value])

@loop_pattern_func
def PSum(n, total, **kwargs):
    """
    Returns a Pattern of length 'n' that sums to equal 'total'

    e.g. PSum(3,8) -> P[3, 3, 2]
         PSum(5,4) -> P[1, 0.75, 0.75, 0.75, 0.75]
    """
    lim = kwargs.get("lim", 0.125)

    data = [total + 1]

    step = 1
    while sum(data) > total:
        data = [step for x in range(n)]
        step *= 0.5

    i = 0
    while sum(data) < total and step >= lim:
        if sum(data) + step > total:
            step *= 0.5
        else:
            data[i % n] += step
            i += 1

    return Pattern(data)

@loop_pattern_func
def PRange(start, stop=None, step=1):
    """ Returns a Pattern equivalent to ``Pattern(range(start, stop, step))`` """
    if stop is None:
        
        start, stop = 0, start

    if start == stop:

        return Pattern(start)
    
    if (start > stop and step > 0) or (start < stop and step < 0):
    
        step = step*-1

    return Pattern(list(range(start, stop, step)))

@loop_pattern_func
def PTri(start, stop=None, step=1):
    """ Returns a Pattern equivalent to ``Pattern(range(start, stop, step))`` with its reversed form
    appended."""
    if stop is None: 
        start, stop = 0, start
    pat = PRange(start, stop, step)
    return pat | pat.reverse()[1:-1]

@loop_pattern_func
def PSine(n=16):
    """ Returns values of one cycle of sine wave split into 'n' parts """
    i = (2 * math.pi) / n
    return Pattern([math.sin(i * j) for j in range(int(n))])

@loop_pattern_func
def PEuclid(n, k):
    ''' Returns the Euclidean rhythm which spreads 'n' pulses over 'k' steps as evenly as possible.
        e.g. `PEuclid(3, 8)` will return `P[1, 0, 0, 1, 0, 0, 1, 0]` '''
    return Pattern( EuclidsAlgorithm(n, k) )

@loop_pattern_func
def PEuclid2(n, k, lo, hi):
    ''' Same as PEuclid except it returns an array filled with 'lo' value instead of 0
        and 'hi' value instead of 1. Can be used to generate characters patterns used to
        play sample like play(PEuclid2(3,8,'-','X')) will be equivalent to
        play(P['X', '-', '-', 'X', '-', '-', 'X', '-'])
        that's like saying play("X--X--X-")
        '''
    return Pattern( EuclidsAlgorithm(n, k, lo, hi) )

@loop_pattern_func
def PBern(size=16, ratio=0.5):
    """ Returns a pattern of 1s and 0s based on the ratio value (between 0 and 1).
        This is called a Bernoulli sequence. """
    return Pattern([int(random.random() < ratio) for n in range(size)])


def PBeat(string, start=0, dur=0.5):
    """ Returns a Pattern of durations based on an input string where
        non-whitespace denote a pulse e.g.
        ::

            >>> PBeat("x xxx x")
            P[1, 0.5, 0.5, 1, 0.5]
    """
    data = [int(char != " ") for char in list(string)]
    pattern = Pattern(PulsesToDurations( data ))
    if start != 0:
        pattern = pattern.rotate(int(start))
    return pattern * dur

@loop_pattern_func
def PDur(n, k, start=0, dur=0.25):
    """ Returns the *actual* durations based on Euclidean rhythms (see PEuclid) where dur
        is the length of each step.
        ::
            >>> PDur(3, 8)
            P[0.75, 0.75, 0.5]
            >>> PDur(5, 16)
            P[0.75, 0.75, 0.75, 0.75, 1]
    """
    # If we have more pulses then steps, double the steps and decrease the duration

    while n > k:
        k   = k * 2
        dur = dur / 2

    pattern = Pattern(PulsesToDurations( EuclidsAlgorithm(n, k) ))
    if start != 0:
        pattern = pattern.rotate(int(start))
    return pattern * dur

@loop_pattern_func # forces it into a stream instead of Group
def PDelay(*args):
    return PDur(*args).accum().group()

@loop_pattern_func
def PStrum(n=4):
    """ Returns a pattern of durations similar to how you might strum a guitar """
    return (Pattern([1,1/2]).stutter([1,n + 1])|Pattern([1.5,1/2]).stutter([1,n])|1)

def PQuicken(dur=1/2, stepsize=3, steps=6):
    """ Returns a PGroup of delay amounts that gradually decrease """
    delay = []
    count = 0
    for i in range(steps):
        for j in range(stepsize-1):
            delay.append( count )
            count += dur / stepsize
        dur /= stepsize
    return PGroup(delay)


def PRhythm(durations):
    """ Converts all tuples/PGroups into delays calculated using the PDur algorithm.
    e.g.
        PRhythm([1,(3,8)]) -> P[1,(2,0.75,1.5)]
    *work in progress*
    """
    if len(durations) == 1:
        item = durations[0]
        if isinstance(item, (tuple, PGroup)):
            return PDur(*item)
        else:
            return durations
    else:
        durations = asStream(durations).data
        output = Pattern()
        for i in range(len( asStream(durations).data ) ):
            item = durations[i]
            if isinstance(item, (list, Pattern)):
                value = PRhythm(item)
            elif isinstance(item, (tuple, PGroup)):
                dur   = PDur(*item)
                value = PGroup(sum(dur)|dur.accum()[1:])
            else:
                value = item
            output.append(value)
        return output


def PJoin(patterns):
    """ Joins a list of patterns together """
    data = []
    for pattern in patterns:
        data.extend(pattern)
    return Pattern(data)



from __future__ import absolute_import, division, print_function

from .Code import WarningMsg
from .Patterns import Pattern, Cycle, asStream
from .Utils import modi
from .TimeVar import var, Pvar

import inspect

class MethodList:
    """ Class for holding information about the order of which methods have been
        called on Player attributes. `root` is the original Pattern.
     """
    def __init__(self, root):
        self.root=root
        self.list_of_methods = []

    def get_root_pattern(self):
        return self.root

    def set_root_pattern(self, new):
        self.root = new

    def add_method(self, method_name, args, kwargs):
        self.list_of_methods.append((method_name, args, kwargs))

    def update(self, method, args, kwargs):
        """ Updates the args and kwargs for a repeated method """
        for i, info in enumerate(self.list_of_methods):
            name, _, _  = info
            if name == method:
                self.list_of_methods[i] = (method, args, kwargs)
                return
        raise ValueError

    def remove(self, method):
        """ Removes a method (should be a string) from the list of methods """
        for i, info in enumerate(self.list_of_methods):
            name, args, kwargs = info
            if name == method:
                self.list_of_methods.pop(i)
                return
        raise ValueError

    def __repr__(self):
        return repr(self.list_of_methods)

    def __contains__(self, method):
        """ Returns true if the method is in the list of methods """
        for name, args, kwargs in self.list_of_methods:
            if name == method:
                return True
        return False

    def __iter__(self):
        for value in self.list_of_methods:
            yield value

class Repeatable(object):
    after_update_methods = []
    method_synonyms      = {}
    def __init__(self):
        self.repeat_events        = {}
        self.previous_patterns    = {} # not a good name - TODO change

    def update_pattern_root(self, attr):
        """ Update the base attribute pattern that methods are applied to """

        if attr not in self.previous_patterns:

            self.previous_patterns[attr] = MethodList(self.attr[attr])

        else:

            self.previous_patterns[attr].set_root_pattern( self.attr[attr] )

        self.update_pattern_methods(attr)     

        return

    def update_pattern_methods(self, attr):
        """ Update the 'current' version of a pattern based on its root and methods stored """

        if attr not in self.previous_patterns:

            self.previous_patterns[attr] = MethodList(self.attr[attr])

        result = self.previous_patterns[attr].get_root_pattern()

        # For each method in the list, call on the pattern

        for method, args, kwargs in self.previous_patterns[attr]:

            call_pattern_method = getattr(Pattern, method)

            result = call_pattern_method(result, *args, **kwargs)

        self.attr[attr] = result

        return

    def get_attr_and_method_name(self, cmd):
        """ Returns the attribute and method name from a string in the form
            `"attr.method"` would return `"attr"` and `"method"`. If attr is not
            present, it returns `"degree"` in place. 
        """

        if cmd in self.method_synonyms:

            attr = [ self.method_synonyms[cmd] ]

        else:

            attr = cmd.split(".")

        # We can also schedule attribute methods

        if len(attr) == 1:

            attr_name   = "degree"
            method_name = attr[0]

        elif len(attr) == 2:

            attr_name = attr[0]
            method_name = attr[1]

        return attr_name, method_name

    def is_pattern_method(self, method_name, attr="degree"):
        """ Returns True if the method is a valid method of `Pattern` """

        if attr == "degree" and hasattr(self, method_name):

            return False

        elif hasattr(Pattern, method_name):

            return True

        else:

            return False

    def is_player_method(self, method_name, attr="degree"):
        """ Returns True if the method is a valid method  of `Player` """ 
        return hasattr(self, method_name) and attr == "degree"

    def get_method_by_name(self, cmd):
        """ Returns the attribute name and method based on `cmd` which is a string.
            Should be in form `"attr.method"`.
        """

        attr_name, method_name = self.get_attr_and_method_name(cmd)

        # Just get the player method if a valid player method

        if self.is_player_method(method_name, attr_name):

            method = getattr(self, method_name)

        # If its a Pattern method, create a "new" function  that acts as a method

        elif self.is_pattern_method(method_name, attr_name):

            def method(*args, **kwargs):

                # If there are no "old" patterns held in memory, use the pattern method and store

                if attr_name not in self.previous_patterns:

                    self.previous_patterns[attr_name] = MethodList(self.attr[attr_name]) # store the root

                # If this has already been called, "undo it"

                if method_name in self.previous_patterns[attr_name]:

                    self.previous_patterns[attr_name].remove(method_name)

                # If not, add it to the list

                else:

                    self.previous_patterns[attr_name].add_method(method_name, args, kwargs)

                # Update the attribute

                self.update_pattern_methods(attr_name)

                return

            method.__name__ = cmd # for debugging purposes

        else:

            # Raise TypeError if not a method

            err = "{} is not a valid method for type {}".format(cmd, self.__class__)
            
            raise(TypeError(err))

        assert callable(method)
        
        return attr_name, method

    def after(self, n, cmd, *args, **kwargs):
        """ Schedule self.cmd(args, kwargs) in 'n' beats time
            ```
            # Stop the player looping after 16 beats
            p1 >> pads().after(16, "stop")
            ```
        """
        quantise = kwargs.get("quantise", True)

        try:

            event = lambda: getattr(self, cmd)(*args, **kwargs)

            if not quantise:

                time = self.metro.now() + n

            else:

                time = self.metro.next_bar() + n

            self.metro.schedule( event, time )

        except:

            pass
        
        return self
        
    def every(self, occurence, cmd, *args, **kwargs):
        """ Every n beats, call a method (defined as a string) on the
            object and use the args and kwargs. To call the method
            every n-th beat of a timeframe, use the `cycle` keyword argument
            to specify that timeframe.

            ::
                # Call the shuffle method every 4 beats

                p1.every(4, 'shuffle')

                # Call the stutter method on the 5th beat of every 8 beat cycle

                p1.every(5, 'stutter', 4, cycle=8)

                # If the method is not valid but *is* a valid Pattern method, that is called and reverted

                p1.every(4, 'palindrome')
        """

        try:

            attr, method = self.get_method_by_name(cmd)

        except AttributeError:

            WarningMsg("{} is not a valid method for type {}".format(cmd, self.__class__))

            return self

        # Collect the cycle length

        cycle = None
        ident = None

        if "cycle" in kwargs:

            cycle = kwargs["cycle"]
            del kwargs["cycle"]

        if "ident" in kwargs:

            ident = kwargs["ident"]
            del kwargs["ident"]

        # Convert `Cycles` to `var`-- should they be Pvar?

        attr_name, method_name = self.get_attr_and_method_name(cmd)

        # Just get the player method if a valid player method

        if self.is_player_method(method_name, attr_name):

            cycle_dur = occurence

        else:

            cycle_dur = occurence * 2

        args, kwargs = self.convert_cycles(args, kwargs, cycle_dur)

        # If the method call already exists, just update it (should be in a function)

        if ident is not None:

            cmd = "{}-{}".format(cmd, ident)

        if cmd in self.repeat_events:

            # Work out whether the method needs calling or not

            call = self.repeat_events[cmd]

            # Update the time details

            call.update(occurence, cycle, args, kwargs)

            attr, method_name = self.get_attr_and_method_name(cmd)

            if self.is_pattern_method(method_name, attr):

                # if n is even, it means the method is active # TODO -- put this in a class mate

                n, acc = call.count()

                if n % 2 == 1:

                    if method_name in self.previous_patterns[attr]:

                        self.previous_patterns[attr].remove(method_name)

                else:

                    if method_name in self.previous_patterns[attr]:

                        self.previous_patterns[attr].update(method_name, args, kwargs)

                # Update the attribute

                self.update_pattern_methods(attr)

            if not call.isScheduled():

                call.schedule()

        else:

            self.repeat_events[cmd] = MethodCall(self, method, occurence, cycle, args, kwargs)

            self.repeat_events[cmd].schedule()

        return self

    def stop_calling_all(self):
        """ Stops all repeated methods. """
        for method in list(self.repeat_events.keys()):

            self.never(method)

        return self            

    def never(self, cmd, ident=None):
        """ Stops calling cmd on repeat """

        attr, method = self.get_attr_and_method_name(cmd)

        if ident is not None:
        
            cmd = "{}-{}".format(cmd, ident)

        try:
            # If it a pattern method, undo it
            
            if method in self.previous_patterns[attr]:
            
                self.previous_patterns[attr].remove(method)
            
                self.update_pattern_methods(attr)
            
            self.repeat_events[cmd].stop()
            
            del self.repeat_events[cmd]
        
        except KeyError:
        
            err = "Player method '{}' not active".format(cmd)
        
            raise KeyError(err)
        
        return self

    @staticmethod
    def convert_cycles(args, kwargs, occurence):
        """ Converts any values that are instances of `Cycle` to a `var` with the 
            same duration as the frequency of the every call (occurrence) """

        args = [(var(value, occurence) if isinstance(value, Cycle) else value) for value in args]
        kwargs = {key: (var(value, occurence) if isinstance(value, Cycle) else value) for key, value in kwargs.items()}

        return args, kwargs

class MethodCall:
    """ Class to represent an object's method call that,
        when called, schedules itself in the future """
    def __init__(self, parent, method, n, cycle=None, args=(), kwargs={}):
        
        self.parent = parent  
        self.method = method

        self.update(n, cycle, args, kwargs)

        self.after_update = False
        self.stopping = False

    def update(self, n, cycle=None, args=(), kwargs={}):
        """ Updates the values of the MethodCall. Re-adjusts the index if cycle has been changed """

        if cycle is not None:

            self.when   = asStream(cycle)
            self.cycle  = asStream(n)

        else:

            self.when   = asStream(n)
            self.cycle  = None

        self.args = args
        self.kwargs = kwargs

        # Check if a method has the _beat_ keyword argument

        if "_beat_" in inspect.getfullargspec(self.method).args:

            self.kwargs["_beat_"] = None

        self.i, self.next = self.count()

        self.offset = float(modi(self.cycle, self.i)) if self.cycle is not None else 0
        
        return self

    def count(self):
        """ Counts the number of times this method would have been called between clock start and now """

        n = 0; acc = 0; dur = 0
        now = float(self.parent.metro.now())

        # Get durations

        durations = self.when # if self.cycle is None else asStream(self.cycle)
        total_dur = float(sum(durations))

        # How much time left to fit remainder in

        try:
       
            acc = now - (now % total_dur)

        except ZeroDivisionError:

            acc = 0

        # n is the index to return for calculating self.when[n]
        # acc is when to start

        n = int(len(durations) * (acc / total_dur))

        if acc != now:

            while True:

                dur = float(durations[n])

                acc += dur
                n   += 1

                if acc >= now:

                    break

        return n, acc

    def __repr__(self):
        return "<Future {}() call of '{}'>".format(self.method.__name__, self.parent)

    def __call__(self, *args, **kwargs):
        """ Proxy for parent object __call__, calls the enclosed method and schedules it in the future. """

        assert self.method is not None

        # Return without scheduling if stopping
        
        if self.stopping:

            return

        # Give the method a reference to when OSC messages should send

        self.assign_beat()

        # Call the method

        self.call_method()

         # Update the next time to schedule

        self.next += float(self.when[self.i])

        if self.cycle is not None:
            
            self.offset = float(modi(self.cycle, self.i))

        # Re-schedule the method call

        self.schedule()

        # Increase the index to get the next duration
        
        self.i += 1

        return

    def assign_beat(self):
        if "_beat_" in self.kwargs:
            self.kwargs["_beat_"] = self.get_next()
        return

    def call_method(self):
        """ Calls the method. Prints to the console with error info. """
        try:

            self.method.__call__(*self.args, **self.kwargs)

        except Exception as e:

            print("{} in '{}': {}".format(e.__class__.__name__, self.method.__name__, e))

        return

    def get_next(self):
        """ Returns the beat that the next occurrence of this method call is due """
        return self.next + self.offset

    def schedule(self):
        """ Schedules the method to be called in the clock """
        self.parent.metro.schedule(self, self.get_next())

    def isScheduled(self):
        """ Returns True if this is in the Tempo Clock """
        return self in self.parent.metro

    def stop(self):
        self.stopping = True



# Tutoral 12: SynthDefs


# FoxDot creates music by giving player objects a 'digital instrument'
# to play, which are called 'SynthDefs'. You can see the list of pre-installed
# 'Synths' by executing

print(SynthDefs)

# Each one of these represents a `SynthDef` *object*. These objects are then
# given to Players to play - like giving an instrument to someone in your
# orchestra.

# Writing your own Synth Definitions

# This is a bit more advanced, but if you have already written SynthDefs in
# Supercollider then you might feel at home.  If not, come back to this section
# later.

# FoxDot can access any SynthDef stored on the SuperCollider server,
# but it needs to know it's there. If you have already written a SynthDef
# in SuperCollider and named it \mySynth then you just create a SynthDef
# instance using FoxDot like so:

mySynth = SynthDef("mySynth")

# Using the same variable name in FoxDot as in SuperCollider for your SynthDef
# is a good idea to avoid confusion. If you want to write (or edit) your own
# SynthDef during run-time in FoxDot you can use a SuperCollider API by
# importing the SCLang module. All FoxDot SynthDef objects inherit the
# base-class behaviour, such as low- and high-pass filters and vibrato,
# but these can be overridden or updated easily. If you want to know more
# about digital sound processing and SynthDef creation, check out the
# SuperCollider documentation. Below is an example of creating one in FoxDot:

# Import module for writing SCLang code from Python
from SCLang import *

# Create a SynthDef named 'example' (using the same variable name as the SynthDef name is a good idea)
example = SynthDef("example")

# Create the oscillator (osc) using a sine wave
example.osc = SinOsc.ar(ex.freq)

# And give it a percussive sound envelope (env)
example.env = Env.perc()

# Finally, store it!
example.add()

# How to create a SynthDef

with SynthDef("pads") as pads:
	pads.osc = SinOsc.ar(pads.freq)
	pads.env = Env.perc()

# Equivalent to

pads = SynthDef("pads")
pads.osc = SinOsc.ar(pads.freq)
pads.env = Env.perc()
pads.add()



"""
Contains classes `Pattern` and `PGroup` and the base class for `GeneratorPattern` (see Generators.py).
"""

from __future__ import absolute_import, division, print_function

from random import choice, shuffle
from copy import deepcopy

from .Operations import *
from ..Utils import *

import functools
import inspect

# Decorator functions for nested expansion of pattern functions and methods

def loop_pattern_func(f):
    ''' Decorator for allowing any Pattern function to create
        multiple Patterns by using Patterns or TimeVars as arguments '''
    @functools.wraps(f)
    def new_function(*args):
        
        # Return any functions that use TimeVars as PvarGenerators
        timevars = [arg for arg in args if isinstance(arg, Pattern.TimeVar)]
        if len(timevars) > 0:
            return Pattern.TimeVar.CreatePvarGenerator(f, *args)
        
        # Loop the pattern with different values
        pat = Pattern()
        # Force pattern types if using lists/tuples
        args = [PatternFormat(arg) for arg in args]
        # Continually extend the pattern
        for i in range(LCM(*[len(arg) for arg in args if (hasattr(arg, '__len__') and not isinstance(arg, PGroup))])):
            pat |= f(*[(arg[i] if isinstance(arg, Pattern) else arg) for arg in args])
        return pat
    new_function.argspec = inspect.getfullargspec(f)
    return new_function

# TODO -- if it isn't looped, return the original if it is a group

def loop_pattern_method(f):
    ''' Decorator for allowing any Pattern method to create
        multiple (or rather, longer) Patterns by using Patterns as arguments '''
    
    @functools.wraps(f)
    def new_function(self, *args):
        
        # Return any functions that use TimeVars as PvarGenerators
        timevars = [arg for arg in args if isinstance(arg, Pattern.TimeVar)]
        if len(timevars) > 0:
            return Pattern.TimeVar.CreatePvarGenerator(f, *args, pattern=self)

        pat = Pattern()
        # Force pattern types if using lists/tuples
        args = [PatternFormat(arg) for arg in args]
        for i in range(LCM(*[len(arg) for arg in args if (hasattr(arg, '__len__') and not isinstance(arg, PGroup))])):
            pat |= f(self, *[(modi(arg, i) if not isinstance(arg, PGroup) else arg) for arg in args])
        return pat

    new_function.argspec = inspect.getfullargspec(f)
    return new_function

def PatternMethod(f):
    ''' Decorator that makes a function into a metaPattern method'''
    setattr(metaPattern, f.__name__, f)
    return

def StaticPatternMethod(f):
    ''' Decorator that makes a function into a metaPattern static  method'''
    setattr(metaPattern, f.__name__, staticmethod(f))
    return

def ClassPatternMethod(f):
    ''' Decorator that makes a function into a metaPattern class method'''
    setattr(metaPattern, f.__name__, classmethod(f))
    return

# Begin Pattern Abstratct Base Class

class metaPattern(object):
    """ Abstract base class for Patterns """
    WEIGHT = -1
    # data = None
    bracket_style = "[]"
    debugging = False
    meta = []

    def __init__(self, *args):

        if len(args):

            data = args[0]
        
            if type(data) is str:
                
                self.fromString(data)

            elif type(data) is tuple:

                self.data = PGroup(data)
                self.make()

            elif isinstance(data, self.__class__):

                self.data = data.data
                
            else:
                
                self.data = data
                self.make()

        else:

            self.data = []


    def new(self, data):
        """ Returns a new pattern object with this Pattern's class type """
        return self.__class__(data + self.meta)

    def transform(self, func):
        """
        Recursively transforms values and nested patterns
        """
        output = []
        for item in self.data:
            if isinstance(item, (metaPattern, GeneratorPattern)):
                output.append(item.transform(func))
            else:
                output.append(func(item))
        return self.__class__(output)

    def int(self):
        return self.transform(int)

    def float(self):
        return self.transform(float)

    def str(self):
        return self.transform(str)

    @classmethod
    def get_methods(cls):
        """ Returns the methods associated with the `Pattern` class as a list """
        return [attr for attr in dir(cls) if callable(getattr(cls, attr))]

    def get_data(self):
        """ Returns self.data if data is not a single instance of this class, in which 
            case self.data[0].data is returned """
        return self.data

    @classmethod
    def help(cls):
        """ Prints the Pattern class docstring to the console """
        return print(cls.__doc__)
            
    def __len__(self):
        """ Returns the *expanded* length of the pattern such that if the pattern is laced, the
            value is the length of the list multiplied by the lowest-common-multiple of the lengths
            of nested patterns. e.g. the following are identical:
            ```
            >>> print( len(P[0,1,2,[3,4]]) )
            8
            >>> print( len(P[0,1,2,3,0,1,2,4]) )
            8
            ```
        """
        lengths = [1]
        n = 0
        for item in self.data:
            if isinstance(item,  EmptyItem):
                continue
            elif isinstance(item, Pattern):
                lengths.append(len(item))
            n += 1
        return LCM(*lengths) * n

    
    def __str__(self):
        try:
            if len(self.data) > 20:
                val = self.data[:8] + [dots()] + self.data[-8:]
            else:
                val = self.data
        except AttributeError:
            val = self.data
        return "P" + self.bracket_style[:-1] + ( repr(val)[1:-1] ) + self.bracket_style[-1]

    def __repr__(self):
        return str(self)

    # Conversion methods

    def string(self):
        """ Returns a PlayString in string format from the Patterns values """
        string = ""
        for item in self.data:
            if isinstance(item, (PGroup, GeneratorPattern)):
                string += item.string()
            elif isinstance(item, Pattern):
                string += "(" + "".join([(s.string() if hasattr(s, "string") else str(s)) for s in item.data]) + ")"
            else:
                string += str(item)
        return string

    def asGroup(self):
        """ Returns the Pattern as a PGroup """
        return PGroup(self.data)

    def group(self):
        """ Returns the Pattern as a PGroup """
        return PGroup(self.data)

    # TODO -- this is super hacky vv

    def convert_data(self, dtype=float, *args, **kwargs):
        """ Makes a true copy and converts the data to a given data type """
        new = map((lambda x: x.convert_data(dtype, *args, **kwargs) if isinstance(x, metaPattern) else dtype(x, *args, **kwargs)), self.data)
        return self.true_copy(list(new))

    def copy(self):
        """ Returns a copy of the Pattern such that alterations to the
            Pattern.data do not affect the original.
        """
        return self.new(self.data[:])

    def true_copy(self, new_data=None):
        """ Returns a copy of the Pattern such that items within the
            Pattern hold the same state as the original.
        """
        new = self.__class__()
        new.__dict__ = self.__dict__.copy()
        if new_data is not None:
            new.data = new_data
        return new
    
    # Pattern container methods
 
    def __getitem__(self, key):
        """ Calls self.getitem(). Is overridden in `FoxDot.lib.TimeVar`
            for indexing with TimeVars """
        return self.getitem(key)

    def getitem(self, key, get_generator=False):
        """ Called by __getitem__() """
        # We can get multiple values by indexing with a pattern or tuple
        if isinstance(key, (metaPattern, tuple)):
            val = self.new([self.getitem(n) for n in key])
        # We can get items using a slice
        elif isinstance(key, slice):
            val = self.getslice(key.start,  key.stop, key.step)
        else:
            # Get the "nested" single value
            i = key % len(self.data)
            val = self.data[i]
            if isinstance(val, (Pattern, Pattern.Pvar)) or ( isinstance(val, GeneratorPattern) and not get_generator ):
                j   = key // len(self.data)
                val = val.getitem(j, get_generator)
            elif isinstance(val, GeneratorPattern) and get_generator:
                return val
        return val
    
    def __setitem__(self, key, value):
        if isinstance(key, slice):
            self.data[key] = Format(value) # TODO - make sure this works
        else:
            i = key % len(self.data)
            if isinstance(self.data[i], metaPattern):
                j = key // len(self.data)
                self.data[i][j] = value
            else:
                if key >= len(self.data):
                    self.data[i] = Pattern([self.data[i], Format(value)]).stutter([key // len(self.data) , 1])
                else:
                    self.data[i] = Format(value)
        return

    def setitem(self, key, value):
        self.data[key] = Format(value)
            
    def __iter__(self):
        """ Returns a generator object for this Pattern """
        for i in range(len(self)):
            yield self.getitem(i)

    def items(self):
        """ Returns a generator object equivalent to using enumerate() """
        for i, value in enumerate(self):
            yield i, value

    def getslice(self, start, stop, step=1):
        """ Called when using __getitem__ with slice notation """
        start = start if start is not None else 0
        stop  = stop if stop is not None else len(self)
        step  = step if step is not None else 1

        if stop < start:

            stop = (len(self.data) +  stop)

        return Pattern([self[i] for i in range(start, stop, step) ])
            
    def __setslice__(self, i, j, item):
        """ Only works in Python 2 - maybe get rid? """
        self.data[i:j] = Format(item)

    # Integer returning
    
    def count(self, item):
        """ Returns the number of occurrences of item in the Pattern"""
        return self.data.count(item)

    def __add__(self, other):  
        if isinstance(other, GeneratorPattern):
            return other.__radd__(self)
        return PAdd(self, other)

    def __radd__(self, other):  
        if isinstance(other, GeneratorPattern):
            return other.__add__(self)
        return PAdd(self, other)

    def __sub__(self, other):  
        if isinstance(other, GeneratorPattern):
            return other.__rsub__(self)
        return PSub(self, other)
    
    def __rsub__(self, other):  
        if isinstance(other, GeneratorPattern):
            return other.__sub__(self)
        return PSub2(self, other)

    def __mul__(self, other):    
        if isinstance(other, GeneratorPattern):
            return other.__rmul__(self)
        return PMul(self, other)

    def __rmul__(self, other):   
        if isinstance(other, GeneratorPattern):
            return other.__mul__(self)
        return PMul(self, other)

    def __truediv__(self, other):    
        if isinstance(other, GeneratorPattern):
            return other.__rtruediv__(self)
        return PDiv(self, other)

    def __rtruediv__(self, other):    
        if isinstance(other, GeneratorPattern):
            return other.__truediv__(self)
        return PDiv2(self, other)

    def __floordiv__(self, other):  
        if isinstance(other, GeneratorPattern):
            return other.__rfloordiv__(self)
        return PFloor(self, other)

    def __rfloordiv__(self, other):   
        if isinstance(other, GeneratorPattern):
            return other.__floordiv__(self)
        return PFloor2(self, other)

    def __mod__(self, other):  
        if isinstance(other, GeneratorPattern):
            return other.__rmod__(self)
        return PMod(self, other)

    def __rmod__(self, other):  
        if isinstance(other, GeneratorPattern):
            return other.__mod__(self)
        return PMod2(self, other)

    def __pow__(self, other):  
        if isinstance(other, GeneratorPattern):
            return other.__rpow__(self)
        return PPow(self, other)

    def __rpow__(self, other):  
        if isinstance(other, GeneratorPattern):
            return other.__pow__(self)
        return PPow2(self, other)

    def __xor__(self, other):   
        if isinstance(other, GeneratorPattern):
            return other.__rxor__(self)
        return PPow(self, other)

    def __rxor__(self, other):   
        if isinstance(other, GeneratorPattern):
            return other.__xor__(self)
        return PPow2(self, other)

    def __abs__(self):
        return self.new([abs(item) for item in self])

    def __bool__(self):
        """ Returns True if *any* value in the Pattern are greater than zero """
        # NOTE: this used to be ALL 
        return all([bool(item > 0) for item in self])

    def __nonzero__(self):
        return self.__bool__()
    
    def abs(self):
        return abs(self)

    def __invert__(self):
        """ Using the ~ symbol as a prefix to a Pattern will reverse it.
            >>> a = P[:4]
            >>> print(a, ~a)
            P[0, 1, 2, 3], P[3, 2, 1, 0]
        """
        return self.mirror()

    # Piping patterns together using the '|' operator
    
    def __or__(self, other):
        """ Use the '|' symbol to 'pipe' Patterns into on another """
        return self.concat(other)

    def __ror__(self, other):
        """ Use the '|' symbol to 'pipe' Patterns into on another """
        return asStream(other).concat(self)

    # Zipping patterns together using the '&' operator

    def __and__(self, other):
        return self.zip(other)

    def __rand__(self, other):
        return asStream(other).zip(self)
    
    #  Comparisons --> this might be a tricky one
    def __eq__(self, other):
        return PEq(self, other)
    def __ne__(self, other):
        return PNe(self, other)
    def eq(self, other):
        return self.new([int(value == modi(asStream(other), i)) for i, value in enumerate(self)])
    def ne(self, other):
        return self.new([int(value != modi(asStream(other), i)) for i, value in enumerate(self)])
    # def gt(self, other):
    #     return self.__class__([int(value > modi(asStream(other), i)) for i, value in enumerate(self)])
    # def lt(self, other):
    #     return self.__class__([int(value < modi(asStream(other), i)) for i, value in enumerate(self)])
    # def ge(self, other):
    #     return self.__class__([int(value >= modi(asStream(other), i)) for i, value in enumerate(self)])
    # def le(self, other):
    #     return self.__class__([int(value <= modi(asStream(other), i)) for i, value in enumerate(self)])
    def __gt__(self, other):
        #return self.__class__([int(value > modi(asStream(other), i)) for i, value in enumerate(self)])
        values = []
        other = asStream(other)
        for i, value in enumerate(self): # possibly LCM in future
            value = value > other[i]
            if not isinstance(value, PGroup):
                value = int(value)
            values.append(value)
        return self.new(values)

    def __ge__(self, other):
        #return self.__class__([int(value >= modi(asStream(other), i)) for i, value in enumerate(self)])
        values = []
        other = asStream(other)
        for i, value in enumerate(self): # possibly LCM in future
            value = value >= other[i]
            if not isinstance(value, PGroup):
                value = int(value)
            values.append(value)
        return self.new(values)

    def __lt__(self, other):
        #return self.__class__([int(value < modi(asStream(other), i)) for i, value in enumerate(self)])
        values = []
        other = asStream(other)
        for i, value in enumerate(self): # possibly LCM in future
            value = value < other[i]
            if not isinstance(value, PGroup):
                value = int(value)
            values.append(value)
        return self.new(values)

    def __le__(self, other):
        #return self.__class__([int(value <= modi(asStream(other), i)) for i, value in enumerate(self)])
        values = []
        other = asStream(other)
        for i, value in enumerate(self): # possibly LCM in future
            value = value <= other[i]
            if not isinstance(value, PGroup):
                value = int(value)
            values.append(value)
        return self.new(values)
        
    # Methods that return augmented versions of original

    def shuffle(self, n=1):
        """ Returns a new Pattern with shuffled contents. Note: nested patterns
            stay together. To shuffle the contents of nested patterns, use
            `deep_shuffle` or `true_shuffle`.
        """
        items = []

        for i in range(n):
            data = self.data[:]
            shuffle(data)
            items.extend(data)
        return self.new(items)

    def deep_shuffle(self, n=1):
        """ Returns a new Pattern with shuffled contents and shuffles
            any nested patterns. To shuffle the contents of nested patterns
            with the rest of the Pattern's contents, use `true_shuffle`.
        """
        items = []
        for i in range(n):
            data = [(item if not isinstance(item, metaPattern) else item) for item in self.data[:]]
            shuffle(data)
            items.extend(data)
        return self.new(items)

    def true_shuffle(self, n=1):
        """ Returns a new Pattern with completely shuffle contents such
            that nested Patterns are shuffled within the larger Pattern
        """
        items = []
        for i in range(n):
            data = list(self)
            shuffle(data)
            items.extend(data)
        return self.new(items)

    def reverse(self):
        """ Reverses the contents of the Pattern. Nested patterns are
            not reversed. To reverse the contents of nester patterns
            use `Pattern.mirror()`
        """
        new = self.new(self.data[:])
        new.data.reverse()
        return new

    def sort(self, *args, **kwargs):
        """ Used in place of sorted(pattern) to force type """
        return self.new(sorted(self.data, *args, **kwargs))

    def mirror(self):
        """ Reverses the pattern. Differs to `Pattern.reverse()` in that
            all nested patters are also reversed. """
        new = []
        for i in range(len(self.data), 0, -1):

            value = self.data[i-1]

            if hasattr(value, 'mirror'):

                value = value.mirror()
            
            new.append(value)
            
        return self.new(new)

    def stutter(self, n=2, strict=False):
        """ 
        Returns a new Pattern with each item repeated by `n`. Use
        a list of numbers for stutter different items by different
        amount. e.g.
        ```
        >>> P[0, 1, 2, 3].stutter([1,3])
        P[0, 1, 1, 1, 2, 3, 3, 3]
        ```
        Use strict=True to force generator patterns to return the
        same value `n` times in a row.
        """
        n = asStream(n)
        lrg = max(len(self.data), len(n))
        new = []
        for i in range(lrg):
            for j in range(modi(n,i)):
                item = modi(self.data,i)
                if strict and isinstance(item, GeneratorPattern):
                    item = item.copy()
                new.append(item)
        return self.new(new)

    def arp(self, arp_pattern):
        """ Return a new Pattern with each item repeated len(arp_pattern) times
            and incremented by arp_pattern. Useful for arpeggiating. e.g.
            ```
            >>> P[0, 1, 2, 3].arp([0, 2])
            P[0, 2, 1, 3, 2, 4, 3, 5]
            ```
        """
        return self.stutter(len(arp_pattern)) + arp_pattern

    def splice(self, seq, *seqs):
        """ Takes at least list / Pattern and creates a new Pattern by
            adding a value from each pattern in turn to the new pattern.
            e.g.
            ```
            >>> P[0,1,2,3].splice([4,5,6,7],[8,9])
            P[0,4,8,1,5,9,2,6,8,3,7,9]
            ```
        """
        sequences = (self, asStream(seq)) + tuple(asStream(s) for s in seqs)
        size = LCM(*[len(s) for s in sequences])
        new = []
        for i in range(size):
            for seq in sequences:
                new.append(modi(seq, i))
        return self.new(new)

    def invert(self):
        """ Inverts the values with the Pattern.
        """
        new = []
        lrg = float(max(self.data))
        for item in self.data:
            try:
                new.append(item.invert())
            except:
                new.append((((item / lrg) * -1) + 1) * lrg)
        return self.new(new)

    def shufflets(self, n):
        """ Returns a Pattern of 'n' number of PGroups made from shuffled
            versions of the original Pattern """
        new = self.data[:]
        return Pattern([Pattern(new).shuffle().asGroup() for i in range(n)])

    # Loop methods

    @loop_pattern_method
    def pivot(self, i):
        """ Mirrors and rotates the Pattern such that the item at index 'i'
            is in the same place """
        if len(self) > 0:
            mid = len(self) / 2
            if i > mid:
                i = len(self) - i - 1
                new = self.mirror().rotate((2*(i % len(self)))+1)
            else:
                new = self.rotate((2*(i % len(self)))+1).mirror()
        else:
            new = self.copy()
        return new

    @loop_pattern_method
    def accum(self, n=None):
        """ Returns a Pattern that is equivalent to list of sums of that
            Pattern up to that index."""
        if n is None:
            n = len(self)
        new = [0]
        for i in range(n-1):
            new.append( new[-1] + self[i] )
        return self.new(new)

    @loop_pattern_method
    def stretch(self, size):
        """ Stretches (repeats) the contents until len(Pattern) == size """
        new = []
        for n in range(size):
            new.append( modi(self.data, n) )
        new = self.new(new)
        return new

    @loop_pattern_method
    def trim(self, size):
        """ Shortens a pattern until it's length is equal to size - cannot be greater than the length of the current pattern  """
        new = []
        for n in range(min(len(self), size)):
            new.append( modi(self.data, n) )
        new = self.new(new)
        return new

    @loop_pattern_method
    def ltrim(self, size):
        """ Like trim but removes items from the start of the pattern"""
        new = []
        data = self.mirror().data
        for n in range(min(len(self), size)):
            new.append( modi(data, n) )
        new = self.new(new).mirror()
        return new

    @loop_pattern_method
    def loop(self, n, f=None):
        """ Repeats this pattern n times """
        assert n > 0, ".loop() parameter must be greater than 0"
        new = values = list(self)
        for i in range(n - 1):
            if callable(f):
                values = [f(x) for x in values]
            new += list(values)
        return self.new(new)

    @loop_pattern_method
    def duplicate(self, n):
        """ Repeats this pattern n times but keep nested pattern values """
        new = []
        for i in range(n):
            new += self.data
        return self.new(new)

    @loop_pattern_method
    def iter(self, n):
        """ Repeats this pattern n times but doesn't take nested pattern into account for length"""
        return self[:len(self.data)*n]
        #new = []
        #for i in range(len(self.data) * n):
        #    new += self[i]
        #return self.__class__(new)

    @loop_pattern_method
    def swap(self, n=2):
        new = []
        for pair in [list(val) for val in [reversed(self[i:i+n]) for i in range(0, len(self), n)]]:
            for item in pair:
                new.append(item)
        return self.new(new)

    @loop_pattern_method
    def rotate(self, n=1):
        n = int(n)
        new = self.data[n:] + self.data[0:n]
        return self.new(new)

    @loop_pattern_method
    def sample(self, n):
        """ Returns an n-length pattern from a sample"""
        return self.new(random.sample(list(self), n))

    @loop_pattern_method
    def palindrome(self, a=0, b=None):
        """ Returns the original pattern with mirrored version of itself appended.
            a removes values from the middle of the pattern, if positive.
            b removes values from the end of the pattern, should be negative.

            e.g.

            >>> P[:4].palindrome()
            P[0, 1, 2, 3, 3, 2, 1, 0]
            >>> P[:4].palindrome(1)
            P[0, 1, 2, 3, 2, 1, 0]
            >>> P[:4].palindrome(-1)
            P[0, 1, 2, 3, 3, 2, 1]
            >>> P[:4].palindrome(1,-1)
            P[0, 1, 2, 3, 2, 1]

        """
        a = int(a)

        if a < 0:

            a, b = 0, a
        
        return self | self.mirror()[a:b]

    def alt(self, other):
        """ Returns Pattern(other) """
        return self.__class__(other)

    def norm(self):
        """ Returns the pattern with all values between 0 and 1 """
        pos = self - min(self)
        return pos / max(pos)

    def undup(self):
        """ Removes any consecutive duplicate numbers from a Pattern """
        new = []
        last_val = None
        for value in self:
            if value != last_val:
                new.append(value)
            last_val = value                
        return self.new(new)

    def add(self, other):
        return self + other

    @loop_pattern_method
    def limit(self, func, value):
        """ Returns a new Pattern generated by adding elements from
            this Pattern to a new list and repeatedly calling
            `func()` on this list until `func(l)` is greater than `value`
            e.g.
            ```
            >>> print( P[0, 1, 2, 3].limit(sum, 10) )
            P[0, 1, 2, 3, 0, 1, 2]
            ```
        """
        new = []
        i = 0
        while func(new) < value:
            new.append(self[i])
            i+=1
        return self.new(new)

    # Methods that take a non number / pattern argument

    def replace(self, sub, repl):
        """ Replaces any occurrences of "sub" with "repl" """
        new = []
        for item in self.data:
            if isinstance(item, metaPattern):
                new.append(item.replace(sub, repl))
            elif item == sub:
                new.append(repl)
            else:
                new.append(item)
        return self.new(new)

    def submap(self, mapping):
        """ Similar to Pattern.replace, but takes a dictionary of values """
        new = []
        for item in self.data:
            if isinstance(item, metaPattern):
                new.append(item.submap(mapping))
            else:
                new.append(mapping.get(item, item))
        return self.new(new)

    def compress(self, selector):
        """ Removes values from the pattern if the same index in selector is 0. 
            Similar to .select() but maximum length of the new Pattern is the 
            length of the initial Pattern.  """
        s = asStream(selector)
        return self.new([self[i] for i in range(len(self)) if s[i]])

    def select(self, selector):
        """ Removes values from the pattern if the same index in selector is 0  """
        s = asStream(selector)
        # Don't do anything if all values are 1
        if all([value == 1 for value in s]):
            return self
        return self.new([self[i] for i in range(LCM(len(self), len(s))) if s[i]])    

    def layer(self, method, *args, **kwargs):
        """ Zips a pattern with a modified version of itself. Method argument
            can be a function that takes this pattern as its first argument,
            or the name of a Pattern method as a string. """
        
        if callable(method):
            #func = method
            #args = [self.data] + list(args)
            #func = 
            return self.zip(list(map(method, self.data)))
        else:
            func = getattr(self, method)
            assert callable(func)
            return self.zip(func(*args, **kwargs))

    def every(self, n, method, *args, **kwargs):
        """ Returns the pattern looped n-1 times then appended with
            the version returned when method is called on it. """
        return self.loop(n-1).concat(getattr(self, method).__call__(*args, **kwargs))

    def map(self, func):
        """ Returns a Pattern that calls `func` on each item """
        return self.new([(item.map(func) if isinstance(item, metaPattern) else func(item)) for item in self.data])       
    
    def extend(self, seq):
        """ Should return None """
        self.data.extend(map(convert_nested_data, seq))
        return

    def append(self, item):
        """ Converts a new item to PGroup etc and appends """
        self.data.append(convert_nested_data(item))
        return
    
    def i_rotate(self, n=1):
        self.data = self.data[n:] + self.data[0:n]
        return self

    def i_reverse(self):
        self.data.reverse()
        return self

    def i_sort(self):
        self.data = Pattern(sorted(self.data))
        return self

    def i_shuf(self):
        shuffle(self.data)
        return self

    def set(self, index, value):
        self.data[index] = asStream(value)
        return self

    # Boolean tests

    def startswith(self, prefix):
        """ Returns True if the first item in the Pattern is equal to prefix """
        return self.data[0] == prefix
    
    def all(self, func=(lambda x: bool(x))):
        """ Returns true if all of the patterns contents satisfies func(x) - default is nonzero """
        if len(self.data) is 0:
            return False
        
        for item in self.data:
            if not func(item):
                return False
        return True

    # Extension methods
        
    def concat(self, data):
        """ Concatonates this patterns stream with another """
        new = Pattern()
        if isinstance(data, Pattern):
            new.data = self.data + data.data
        elif isinstance(data, (list, str, range)):
            new.data = list(self.data)
            new.data.extend(map(convert_nested_data, data))
        else:
            new.data = list(self.data)
            new.append(data)
        return new

    def zipx(self, other):
        """ Returns a `Pattern` of `PGroups`, where each `PGroup` contains the i-th
            element from each of the argument sequences. The length of the pattern
            is the lowest common multiple of the lengths of the two joining patterns. """
        new = []
        other = asStream(other)
        for i in range(LCM(len(self.data), len(other.data))):
            item1 = self.data[i % len(self.data)]
            item2 = other.data[i % len(other.data)]
            new.append((item1, item2))
        return self.new(new)

    def zip(self, other, dtype=None):
        """ Zips two patterns together. If one item is a tuple, it extends the tuple / PGroup
            i.e. arrow_zip([(0,1),3], [2]) -> [(0,1,2),(3,2)]
        """

        output = Pattern()

        other  = asStream(other)

        dtype = PGroup if dtype is None else dtype

        for i in range(LCM(len(self), len(other))):

            item1 = self.getitem(i, get_generator=True)
            item2 = other.getitem(i, get_generator=True)

            if all([x.__class__== PGroup for x in (item1, item2)]):

                new_item = dtype(item1.data + item2.data)

            elif item1.__class__ == PGroup:

                new_item = dtype(item1.data + [item2])

            elif item2.__class__ == PGroup:

                new_item = dtype([item1] + item2.data)

            else:

                new_item = dtype(item1, item2)

            output.append(new_item)

        return output
    
    def deepzip(self, other):
        new = []
        other = asStream(other)
        for i in range(LCM(len(self.data), len(other.data))):
            p1 = self.data[i % len(self.data)]
            p2 = other.data[i % len(other.data)]
            if isinstance(p1, metaPattern):
                value = p1.deepzip(p2)
            elif isinstance(p2, metaPattern):
                value = p2.deeprzip(p1)
            else:
                value = (p1, p2)
            new.append(value)
        return self.new(new)

    def deeprzip(self, other):
        new = []
        other = asStream(other)
        for i in range(LCM(len(self.data), len(other.data))):
            p1 = self.data[i % len(self.data)]
            p2 = other.data[i % len(other.data)]
            if isinstance(p1, metaPattern):
                value = p1.deeprzip(p2)
            elif isinstance(p2, metaPattern):
                value = p2.deepzip(p1)
            else:
                value = (p2, p1)
            new.append(value)
        return self.new(new)

    # Returns individual elements / slices

    def choose(self):
        """ Returns one randomly selected item """
        return choice(self.data)

    def get_behaviour(self):
        return None

    # Automatic expansion of nested patterns

    def make(self):
        """ This method automatically laces and groups the data """

        #: Force data into an iterable form
        if isinstance(self.data, (str, range)):

            self.data = list(self.data)
            
        elif not isinstance(self.data, PatternType): # not sure about PlayString data
    
            self.data = [self.data]

        self.data = list(map(convert_nested_data, self.data))
            
        # If this only contains a pattern, its redundant to use this as a container
            
        if len(self.data) == 1:

            if isinstance(self.data[0], Pattern):

                self.data = self.data[0].data

            # Replace this pattern with a Pvar if it is the only item in the Pattern itself

            elif isinstance(self.data[0], Pattern.Pvar): # SUPER HACKY

                self.__class__ = self.data[0].__class__
                self.__dict__  = self.data[0].__dict__.copy()
                
        return self

class Pattern(metaPattern):
    """ Base type pattern """
    WEIGHT = 0
    debug = False

class Cycle(Pattern):
    """ Special Case pattern class for cycling values in "every" """
    def __init__(self, *args):
        Pattern.__init__(self, list(args))
    def __str__(self):
        return "Cycle({})".format(Pattern.__str__(self))

class PGroup(metaPattern):
    """
        Class to represent any groupings of notes as denoted by brackets.
        PGroups should only be found within a Pattern object.
        
    """
    WEIGHT = 2
    bracket_style = "()"
    # set this value to negative how many trailing values you don't want treated as "normal"
    ignore = 0

    def __init__(self, seq=[], *args):

        if not args:
            
            if isinstance(seq, metaPattern):

                seq = seq.data

            elif isinstance(seq, tuple):

                seq = list(seq)
        else:

            seq = [seq] + list(args)

        metaPattern.__init__(self, seq)

        # If the PGroup contains patterns, invert it to a Pattern of PGroups
        
        l = [len(p) for p in self.data if isinstance(p, Pattern)]

        if len(l) > 0:

            new_data = []

            for key in range(LCM(*l)):

                new_data.append(self.__class__([item.getitem(key) if isinstance(item, Pattern) else item for item in self.data]))

            self.__class__ = Pattern

            self.data = new_data

    def merge(self, value):
        """ Merge values into one PGroup """
        if hasattr(value, "__len__"):
            new_data = list(value)
        else:
            new_data = [value]
        return self.new(list(self.data) + new_data)

    def flatten(self):
        """ Returns a nested PGroup as un-nested e.g.
        ::

            >>> P(0,(3,5)).flatten()
            P(0, 3, 5)
        """
        values = []
        for item in self:
            if isinstance(item, PGroup):
                values.extend(list(item))
            else:
                values.append(item)
        return PGroup(values)

    def concat(self, data):
        """ Concatonates this patterns stream with another """
        new = PGroup()
        if isinstance(data, PGroup):
            new.data = self.data + data.data
        # Creates a pattern
        elif isinstance(data, Pattern):
            args = list(self.data)
            args.append(data)
            new = PGroup(*args)
        elif isinstance(data, (list, str, range)):
            new.data = list(self.data)
            new.data.extend(map(convert_nested_data, data))
        else:
            new.data = list(self.data)
            new.append(data)
        return new

    def _get_step(self, dur):
        return dur

    def _get_delay(self, delay):
        return 0

    def _get_sample(self):
        return 0

    def calculate_time(self, dur):
        """ Returns a PGroup of durations to use as the delay argument
            when this is a sub-class of `PGroupPrime` """
        values = []
        step  = self._get_step(dur)
        for i, item in enumerate(self):
            delay = self._get_delay( i * step )
            if isinstance(item, PGroup):
                delay += item.calculate_time( step )
            values.append( delay )
        return PGroup(values)

    def calculate_sample(self):
        values = []
        for item in self.data:
            if isinstance(item, PGroup):
                sample = item.calculate_sample()
            else:
                sample = None            
            values.append(sample)
        if all([v is None for v in values]):
            return None
        else:
            return self.__class__(values) # could cause adding issues

    def get_behaviour(self):
        """ Returns a function that changes a player event dictionary """
        def action(event, key):
            this_delay = self.calculate_time(float(event['dur']))
            return self._update_event(event, key, this_delay)
        return action

    def _update_event(self, event, key, delay):
        sample = self.calculate_sample()
        event = self._update_sample(event, sample)
        event = self._update_delay(event, delay)
        return event

    @staticmethod
    def _update_delay(event, delay):
        """ Updates the delay value in the event dictionary """

        event["delay"] = sum_delays(event["delay"], delay)

        return event

    @staticmethod
    def _update_sample(event, sample):
        """ Updates the sample value in the event dictionary """
        if isinstance(sample, PGroup):
            new_sample = sample.replace(None, 0)
            old_sample = event["sample"] * (sample == None)
            event["sample"] = new_sample + old_sample
        elif sample is not None:
            event["sample"] = sample
        return event

    def has_behaviour(self):
        """ Returns True if this is a PGroupPrime or any elements are
            instances of PGroupPrime or its sub-classes"""
        for value in self:
            if isinstance(value, PGroup) and value.has_behaviour():
                return True
        else:
            return False

    def get_name(self):
        return self.__class__.__name__

    def ne(self, other):
        """ Not equals operator """
        values = []
        other  = PatternFormat(other)
        if isinstance(other, Pattern):
            return other.ne(self)
        for i, item in enumerate(self.data): # possibly LCM?
            item = item != modi(other,i)
            if not isinstance(item, metaPattern):
                item = int(item)
            values.append(item)
        #return self.__class__(values)
        return PGroup(values)

    def __ne__(self,  other):
        return self.ne(other)

    def eq(self, other):
        """ equals operator """
        values = []
        other  = PatternFormat(other) # bad function name
        if isinstance(other, Pattern):
            return other.eq(self)
        for i, item in enumerate(self.data): # possibly LCM?
            item = item == modi(other,i)
            if not isinstance(item, metaPattern):
                item = int(item)
            values.append(item)
        # return self.__class__(values)
        return PGroup(values)

    def __hash__(self):
        return hash( self.__key() )

    def __key(self):
        """ Returns a tuple of information to identify this Pattern """
        return (self.__class__, tuple(self.data))

    def __eq__(self, other):
        return self.eq(other)

    def __gt__(self, other):
        values = []
        other  = PatternFormat(other)
        if isinstance(other, Pattern):
            return other < self
        for i, item in enumerate(self): # possibly LCM
            item = item > modi(other,i)
            if not isinstance(item, metaPattern):
                item = int(item)
            values.append(item)
        return self.new(values)

    def __lt__(self, other):
        values = []
        other  = PatternFormat(other)
        if isinstance(other, Pattern):
            return other > self
        for i, item in enumerate(self): # possibly LCM
            item = item < modi(other,i)
            if not isinstance(item, metaPattern):
                item = int(item)
            values.append(item)
        return self.new(values)

    def __ge__(self, other):
        values = []
        other  = PatternFormat(other)
        if isinstance(other, Pattern):
            return other <= self
        for i, item in enumerate(self): # possibly LCM
            item = item >= modi(other,i)
            if not isinstance(item, metaPattern):
                item = int(item)
            values.append(item)
        return self.new(values)

    def __le__(self, other):
        values = []
        other  = PatternFormat(other)
        if isinstance(other, Pattern):
            return other >= self
        for i, item in enumerate(self): # possibly LCM
            item = item <= modi(other,i)
            if not isinstance(item, metaPattern):
                item = int(item)
            values.append(item)
        return self.new(values)

import random

class GeneratorPattern:
    """
        Used for when a Pattern does not generate a set length pattern,
        e.g. random patterns
    """
    MAX_SIZE = 65536
    debugging = False

    def __init__(self, **kwargs):

        # Set the seed if a random pattern

        self.args = tuple()
        self.kwargs = kwargs
            
        self.mod = Pattern()
        self.mod_functions = []
        self.name  = self.__class__.__name__
        self.parent = None
        self.last_value = None
        self.data  = []
        self.index   = 0
        self.cache = {}

    def __repr__(self):
        """ String version is the name of the class and its arguments """
        return "{}({})".format(self.name, self.data)

    @classmethod
    def help(cls):
        return print(cls.__doc__)
        
    def getitem(self, index=None, *args):
        """ Calls self.func(index) to get an item if index is not in
            self.history, otherwise returns self.history[index] """
        if index is None:
            index, self.index = self.index, self.index + 1
        # If we have already accessed by this index, return the value
        if index in self.cache:
            return self.cache[index]
        else:
            # Calculate new value
            value = self.func(index)
            # Store if we refer to the same index
            self.cache[index] = value
            return value

    @property
    def CACHE_HEAD(self):
        ''' Returns the last value used if it exists '''
        return self.cache.get(self.index - 1)

    def new(self, other, func=Nil):
        """ Creates a new `GeneratorPattern` that references
            this pattern but returns a modified value based on
            func. """
        new = GeneratorPattern()
        new.parent = self
        new.name   = new.parent.name
        new.other  = asStream(other) # We want to store the pattern I think?
        new.data   = "{} {}".format(func.__name__, other)
        new.func   = lambda index: func(new.parent.getitem(index), new.other[index])
        return new

    def func(self, index):
        return index

    @staticmethod
    def from_func(pattern_generator_func):
        """ Create a generator which invokes a given function
            to generate items. The given function should take
            and integer argument and return a pattern item. """
        class CustomGeneratorPattern(GeneratorPattern):
            def func(self, index):
                return pattern_generator_func(index)
        return CustomGeneratorPattern()


    def __int__(self):  
        return int(self.getitem())

    def __float__(self):
        return  float(self.getitem())
    
    # Arithmetic operations create new GeneratorPatterns
    def __add__(self, other):
        return self.new(other, Add)
    def __radd__(self, other):
        return self.new(other, Add)
    def __sub__(self, other):
        return self.new(other, Sub)
    def __rsub__(self, other):
        return self.new(other, rSub)
    def __mul__(self, other):
        return self.new(other, Mul)
    def __rmul__(self, other):
        return self.new(other, Mul)
    def __div__(self, other):
        return self.new(other, Div)
    def __truediv__(self, other):
        return self.new(other, Div)
    def __rdiv__(self, other):
        return self.new(other, rDiv)
    def __rtruediv__(self, other):
        return self.new(other, rDiv)
    def __mod__(self, other):
        return self.new(other,  Mod)
    def __rmod__(self, other):
        return self.new(other, rMod)
    # Container methods
    def __iter__(self):
        for i in range(self.MAX_SIZE):
            yield self[i]
            
    def __getitem__(self, key):
        if type(key) is int:
            return self.getitem(key)
        elif type(key) is slice:
            a = key.start if key.start else 0
            b = key.stop
            c = key.step if key.step else 1
            return Pattern([self[i] for i in range(a, b, c)])

    def dup(self, n=2):
        """ Returns a PGroup with n lots of the Generator """
        return PGroup([self.__class__(*self.args, **self.kwargs) for i in range(n)])

    def transform(self, func):
        """ Use func, which should take 1 argument, to transform the values in a generator pattern. Trivial example:
            myGenerator.transform(lambda x: 0 if x in (0,1,2) else 3)
        """
        return self.new(None, lambda a, b: func(a))

    def map(self, mapping, default=0):
        """ Using .transform() to map values via a dictionary

            ::
                a = PRand([0,1])
                b = a.map({0: 16, 1: 25})

        """
        return self.transform( lambda value: mapping.get(value, default) )

    def copy(self):
        '''
        Returns a new Pattern Generator with same inputs
        '''
        return self.__class__(*self.args, **self.kwargs)

        # TODO - handle callables
        # funcs = {}
        
        # for key, value in mapping.items():
            
        #     # We can map using a function
            
        #     if callable(key) and callable(value):

        #         funcs[partial(lambda: key(self.now()))]  = partial(lambda: value(self.now()))

        #     elif callable(key) and not callable(value):

        #         funcs[partial(lambda: key(self.now()))]  = partial(lambda e: e, value)

        #     elif callable(value):

        #         funcs[partial(lambda e: self.now() == e, key)] = partial(lambda: value(self.now()))

        #     else:
        #         # one-to-one mapping
        #         funcs[partial(lambda e: self.now() == e, key)] = partial(lambda e: e, value)

        # def mapping_function(a, b):
        #     for func, result in funcs.items():
        #         if bool(func()) is True:
        #             value = result()
        #             break
        #     else:
        #         value = default
        #     return value

        # new = self.child(0)        
        # new.calculate = mapping_function
        # return new


class PatternContainer(metaPattern):
    def getitem(self, key, *args):
        key = key % len(self)
        return self.data[key]
    def __len__(self):
        return len(self.data)
    def __str__(self):
        return str(self.data)
    def __repr__(self):
        return str(self)

class EmptyItem(object):
    """ Can be used in a pattern and and is essentially not there """
    def __init__(self):
        pass
    def __repr__(self):
        return "_"

"""    Utility functions and data
"""

# Used to force any non-pattern data into a Pattern

PatternType = (Pattern, list)

def asStream(data):
    """ Forces any data into a [pattern] form """
    return data if isinstance(data, Pattern) else Pattern(data)

def PatternFormat(data):
    """ If data is a list, returns Pattern(data). If data is a tuple, returns PGroup(data).
        Returns data if neither. """
    if isinstance(data, list):
        return Pattern(data)
    if isinstance(data, tuple):
        return PGroup(data)
    return data

def PatternInput(data):
    if isinstance(data, GeneratorPattern):
        return data
    return asStream(data)

Format = PatternFormat ## TODO - Remove this

def convert_nested_data(data):
    """ Converts a piece of data in a pattern to a PGroup/Pattern as appropriate """
    from ..Constants import NoneConst

    if isinstance(data, (int, float)):

        return data

    elif data == None:

        return NoneConst()

    elif type(data) is tuple:
        
        return PGroup(data)
    
    elif type(data) is list or (type(data) is str and len(data) > 1):
    
        return Pattern(data)

    else:

        return data

def patternclass(a, b):
    return PGroup if isinstance(a, PGroup) and isinstance(b, PGroup) else Pattern

def Convert(*args):
    """ Returns tuples/PGroups as PGroups, and anything else as Patterns """
    PatternTypes = []
    for val in args:
        if isinstance(val, (Pattern, PGroup)):
            PatternTypes.append(val)
        elif isinstance(val, tuple):
            PatternTypes.append(PGroup(val))
        else:
            PatternTypes.append(Pattern(val))
    return PatternTypes if len(PatternTypes) > 0 else PatternTypes[0]

def asPattern(item):
    if isinstance(item, metaPattern):
        return item
    if isinstance(item, list):
        return Pattern(item)
    if isinstance(item, tuple):
        return PGroup(item)
    return Pattern(item)

def pattern_depth(pat):
    """ Returns the level of nested arrays """	
    total = 1
    for item in pat:
        if isinstance(item, PGroup):
            depth = pattern_depth(item)
            if depth + 1 > total:
                total = depth + 1
    return total

def equal_values(this, that):
    """ Returns True if this == that """
    comp = this == that
    if isinstance(comp, metaPattern):
        return all(list(comp))        
    else:
        return comp

def group_modi(pgroup, index):
    """ Returns value from pgroup that modular indexes nested groups """
    std_type = (int, float, str, bool)
    if isinstance(pgroup, Pattern.TimeVar) and isinstance(pgroup.now(), std_type):
            return pgroup
    elif isinstance(pgroup, std_type):
        return pgroup
    try:
        return group_modi(pgroup[index % len(pgroup)], index // len(pgroup))
    except(TypeError, AttributeError, ZeroDivisionError):
        return pgroup

def get_avg_if(item1, item2, func = lambda x: x != 0):
    if isinstance(item1, PGroup):
        result = item1.avg_if(item2, func)
    elif isinstance(item2, PGroup):
        result = item2.avg_if(item1, func)
    else:
        result = avg_if_func(item1, item2, func)
    return result

def sum_delays(a, b):
    if bool(a == b):
        return a
    
    if not isinstance(a, PGroup):
        
        a = PGroup(a)
    
    if not isinstance(b, PGroup):
        
        b = PGroup(b)

    sml, lrg = sorted((a, b), key=lambda x: len(x))

    if all([item in lrg for item in sml]):

        value = lrg

    else:

        value = a + b

    return value if len(value) > 1 else value[0]


def force_pattern_args(f):
    """ Wrapper for forcing arguments to be a Pattern """
    def new_func(*args, **kwargs):
        new_args = tuple(x if isinstance(x, metaPattern) else PGroup(x) for x in args)
        new_kwargs = {
            key: value if isinstance(value, metaPattern)
            else value 
            for key, value in kwargs.items()
        }
        return f(*new_args, **new_kwargs)
    return new_func


from __future__ import absolute_import, print_function

import os

from .Composer import compose
from .VoiceSpecificator import generateVoiceSpecification
from .Sinsy import download


def renderizeVoice(outputName,lyrics,notes,durations,tempo,scale,sex,foxdot_root):

	# Constants
	print(foxdot_root)
	FILES_ROOT = os.path.realpath(foxdot_root + "/lib/Extensions/VRender/tmp/")
	LAST_MIDI = FILES_ROOT + "/last_midi_generated_by_vrender.mid"
	VOICE_XML_ORIGINAL=FILES_ROOT + "/last_voice.musicxml"
	VOICE_XML_PROCESSED=FILES_ROOT+"/last_voice.xml"

	WAVS_ROOT = os.path.realpath(foxdot_root + "/snd/_loop_/")
	LAST_VOICE_WAV = WAVS_ROOT + "/last_voice_generated.wav"


	print("Running voice renderization")

	compose(notes,durations,scale,LAST_MIDI,VOICE_XML_ORIGINAL)

	lyrics = tokenize(lyrics)

	generateVoiceSpecification(lyrics,tempo,VOICE_XML_ORIGINAL,VOICE_XML_PROCESSED)

	if sex == "male":
		urlfileName = os.popen("curl -X POST -F 'SPKR_LANG=english' -F 'SPKR=5' -F 'SYNALPHA=0.55' -F 'VIBPOWER=1' -F 'F0SHIFT=0' -F  'SYNSRC=@" + VOICE_XML_PROCESSED +"' http://sinsy.sp.nitech.ac.jp/index.php | grep 'lf0'").read()
	else:
		urlfileName = os.popen("curl -X POST -F 'SPKR_LANG=english' -F 'SPKR=4' -F 'SYNALPHA=0.55' -F 'VIBPOWER=1' -F 'F0SHIFT=0' -F  'SYNSRC=@" + VOICE_XML_PROCESSED +"' http://sinsy.sp.nitech.ac.jp/index.php | grep 'lf0'").read()
	download(urlfileName,LAST_VOICE_WAV)

	os.system("cp " + LAST_VOICE_WAV + " " + WAVS_ROOT + "/" + outputName + ".wav")

	print("Finished voice renderization")

def tokenize(text):
	textSyllables = cleanText(text)
	return filter(lambda x: len(x) > 0, textSyllables.replace(" ", "-").split("-"))

def cleanText(text):

	text.replace("\n"," ")
	text = text.lower()

	symbolsToDelete = ".,'!?" + '"'
	for symbol in symbolsToDelete:
		text = text.replace(symbol,"")

	return text